from typing import BinaryIO, Sequence, Dict, Type

from mindsdb.api.mysql.mysql_proxy.classes.sql_statement_parser import SqlStatementParser
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_fields import PostgresField
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message import PostgresMessage
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_message_identifiers import \
    PostgresBackendMessageIdentifier, PostgresFrontendMessageIdentifier, PostgresAuthType

from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_packets import PostgresPacketReader
from mindsdb.api.postgres.postgres_proxy.utilities import strip_null_byte


# All docstrings for Messages are taken from
# https://www.postgresql.org/docs/current/protocol-message-formats.html as of 2023-2-8 for Postgresql 15


class NoticeResponse(PostgresMessage):
    """
    NoticeResponse (B)
    Byte1('N')
    Identifies the message as a notice.

    Int32
    Length of message contents in bytes, including self.

    The message body consists of one or more identified fields, followed by a zero byte as a terminator. Fields can
    appear in any order. For each field there is the following:

    Byte1
    A code identifying the field type; if zero, this is the message terminator and no string follows. The
    presently defined field types are listed in Section 55.8. Since more field types might be added in future,
    frontends should silently ignore fields of unrecognized type.

    String
    The field value. """  # noqa

    def __init__(self):
        self.identifier = PostgresBackendMessageIdentifier.NOTICE_RESPONSE
        self.backend_capable = True
        self.frontend_capable = False
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder().write(write_file=write_file)


class AuthenticationOk(PostgresMessage):
    """
    AuthenticationOk (B)
    Byte1('R')
    Identifies the message as an authentication request.

    Int32(8)
    Length of message contents in bytes, including self.

    Int32(0)
    Specifies that the authentication was successful. """  # noqa

    def __init__(self):
        self.identifier = PostgresBackendMessageIdentifier.AUTHENTICATION_REQUEST
        self.backend_capable = True
        self.frontend_capable = False
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_int32(0) \
            .write(write_file=write_file)


class AuthenticationClearTextPassword(PostgresMessage):
    """
    AuthenticationCleartextPassword (B)
    Byte1('R')
    Identifies the message as an authentication request.

    Int32(8)
    Length of message contents in bytes, including self.

    Int32(3)
    Specifies that a clear-text password is required. """  # noqa

    def __init__(self):
        self.identifier = PostgresBackendMessageIdentifier.AUTHENTICATION_REQUEST
        self.backend_capable = True
        self.frontend_capable = False
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_int32(3) \
            .write(write_file=write_file)


class ReadyForQuery(PostgresMessage):
    """
    ReadyForQuery (B)
    Byte1('Z')
    Identifies the message type. ReadyForQuery is sent whenever the backend is ready for a new query cycle.

    Int32(5)
    Length of message contents in bytes, including self.

    Byte1
     Current backend transaction status indicator. Possible values are 'I' if idle (not in a transaction block);
    'T' if in a transaction block; or 'E' if in a failed transaction block (queries will be rejected until block is
    ended)."""  # noqa

    transaction_status: bytes

    def __init__(self, transaction_status=None):
        self.identifier = PostgresBackendMessageIdentifier.READY_FOR_QUERY
        self.backend_capable = True
        self.frontend_capable = False
        self.transaction_status = transaction_status or b'I'
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_char(self.transaction_status) \
            .write(write_file=write_file)


class CommandComplete(PostgresMessage):
    """
    CommandComplete (B)
    Byte1('C')
    Identifies the message as a command-completed response.

    Int32
    Length of message contents in bytes, including self.

    String
    The command tag. This is usually a single word that identifies which SQL command was completed.

    For an INSERT command, the tag is INSERT oid rows, where rows is the number of rows inserted. oid used to be the object ID of the inserted row if rows was 1 and the target table had OIDs, but OIDs system columns are not supported anymore; therefore oid is always 0.

    For a DELETE command, the tag is DELETE rows where rows is the number of rows deleted.

    For an UPDATE command, the tag is UPDATE rows where rows is the number of rows updated.

    For a SELECT or CREATE TABLE AS command, the tag is SELECT rows where rows is the number of rows retrieved.

    For a MOVE command, the tag is MOVE rows where rows is the number of rows the cursor's position has been changed by.

    For a FETCH command, the tag is FETCH rows where rows is the number of rows that have been retrieved from the cursor.

    For a COPY command, the tag is COPY rows where rows is the number of rows copied. (Note: the row count appears only in PostgreSQL 8.2 and later.) """  # noqa

    tag: bytes

    def __init__(self, tag: bytes):
        self.identifier = PostgresBackendMessageIdentifier.COMPLETE
        self.tag = tag
        self.backend_capable = True
        self.frontend_capable = False
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_string(self.tag) \
            .write(write_file=write_file)


class BindComplete(PostgresMessage):
    """
    BindComplete (B)
    Byte1('2')
    Identifies the message as a Bind-complete indicator.

    Int32(4)
    Length of message contents in bytes, including self. """

    def __init__(self):
        self.identifier = PostgresBackendMessageIdentifier.BIND_COMPLETE
        self.backend_capable = True
        self.frontend_capable = False
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .write(write_file=write_file)


class Error(PostgresMessage):
    """
    ErrorResponse (B)
    Byte1('E')
    Identifies the message as an error.

    Int32
    Length of message contents in bytes, including self.

    The message body consists of one or more identified fields, followed by a zero byte as a terminator. Fields can
    appear in any order. For each field there is the following:

    Byte1 A code identifying the field type; if zero, this is the message terminator and no string follows. The
    presently defined field types are listed in Section 55.8. Since more field types might be added in future,
    frontends should silently ignore fields of unrecognized type.

    String
    The field value. """  # noqa
    severity: bytes
    code: bytes
    message: bytes

    def __init__(self, severity: bytes, code: bytes, message: bytes):
        self.identifier = PostgresBackendMessageIdentifier.ERROR
        self.backend_capable = True
        self.frontend_capable = False
        self.severity = severity
        self.code = code
        self.message = message
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_char(b'S') \
            .add_string(self.severity) \
            .add_char(b'C') \
            .add_string(self.code) \
            .add_char(b'M') \
            .add_string(self.message) \
            .add_char(b'\x00') \
            .write(write_file=write_file)

    @staticmethod
    def from_answer(error_code: bytes, error_message: bytes):
        return Error(severity=b"ERROR", code=error_code, message=error_message)


class ConnectionFailure(Error):
    def __init__(self, message: str = None, charset: str = "UTF-8"):
        if message is None:
            message = "Connection Failure occurred."
        super().__init__(severity="FATAL".encode(encoding=charset), code="08006".encode(encoding=charset),
                         message=message.encode(encoding=charset))


class InvalidSQLStatementName(Error):
    def __init__(self, message: str = None, charset: str = "UTF-8"):
        if message is None:
            message = "Invalid SQL Statement Name"
            super().__init__(severity="FATAL".encode(encoding=charset), code="26000".encode(encoding=charset),
                             message=message.encode(encoding=charset))


class DataException(Error):
    def __init__(self, message: str = None, charset: str = "UTF-8", code: str = "22000"):
        if message is None:
            message = "Data Exception"
            super().__init__(severity="FATAL".encode(encoding=charset), code=code.encode(encoding=charset),
                             message=message.encode(encoding=charset))


class ParameterStatus(PostgresMessage):
    """
    ParameterStatus (B)
    Byte1('S')
    Identifies the message as a run-time parameter status report.

    Int32
    Length of message contents in bytes, including self.

    String
    The name of the run-time parameter being reported.

    String
    The current value of the parameter. """

    def __init__(self, name: bytes, value: bytes):
        self.identifier = PostgresBackendMessageIdentifier.PARAMETER
        self.backend_capable = True
        self.frontend_capable = False
        self.name = name
        self.value = value
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_string(self.name) \
            .add_string(self.value) \
            .write(write_file=write_file)


class RowDescriptions(PostgresMessage):
    """
    RowDescription (B)
    Byte1('T')
    Identifies the message as a row description.

    Int32
    Length of message contents in bytes, including self.

    Int16
    Specifies the number of fields in a row (can be zero).

    Then, for each field, there is the following:

    String
    The field name.

    Int32
    If the field can be identified as a column of a specific table, the object ID of the table; otherwise zero.

    Int16
    If the field can be identified as a column of a specific table, the attribute number of the column; otherwise zero.

    Int32
    The object ID of the field's data type.

    Int16
    The data type size (see pg_type.typlen). Note that negative values denote variable-width types.

    Int32
    The type modifier (see pg_attribute.atttypmod). The meaning of the modifier is type-specific.

    Int16
    The format code being used for the field. Currently will be zero (text) or one (binary). In a
    RowDescription returned from the statement variant of Describe, the format code is not yet known and will always
    be zero."""  # noqa

    fields: Sequence[PostgresField]

    def __init__(self, fields: Sequence[PostgresField]):
        self.identifier = PostgresBackendMessageIdentifier.ROW_DESCRIPTION
        self.backend_capable = True
        self.frontend_capable = False
        self.fields = fields
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .add_int16(len(self.fields)) \
            .add_fields(self.fields) \
            .write(write_file=write_file)


class ParameterDescription(PostgresMessage):
    """
    ParameterDescription (B)
    Byte1('t')
    Identifies the message as a parameter description.

    Int32
    Length of message contents in bytes, including self.

    Int16
    The number of parameters used by the statement (can be zero).

    Then, for each parameter, there is the following:

    Int32
    Specifies the object ID of the parameter data type. """

    def __init__(self, parameters: Sequence):
        self.identifier = PostgresBackendMessageIdentifier.PARAMETER_DESCRIPTION
        self.backend_capable = True
        self.frontend_capable = False
        self.num_params = len(parameters)
        self.parameters = parameters
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        packet = self.get_packet_builder().add_int16(self.num_params)
        for param in self.parameters:
            packet = packet.add_int32(param)
        packet.write(write_file=write_file)


class DataRow(PostgresMessage):
    """
    DataRow (B)
    Byte1('D')
    Identifies the message as a data row.

    Int32
    Length of message contents in bytes, including self.

    Int16
    The number of column values that follow (possibly zero).

    Next, the following pair of fields appear for each column:

    Int32
     The length of the column value, in bytes (this count does not include itself). Can be zero. As a special
    case, -1 indicates a NULL column value. No value bytes follow in the NULL case.

    Byten
    The value of the column, in the format indicated by the associated format code. n is the above length. """  # noqa

    rows: Sequence[Sequence[bytes]]
    num_cols: int

    def __init__(self, rows: Sequence[Sequence[bytes]]):
        self.identifier = PostgresBackendMessageIdentifier.DATA_ROW
        self.backend_capable = True
        self.frontend_capable = False
        if len(rows) != 0:
            self.num_cols = len(rows[0])
        else:
            self.num_cols = 0
        self.rows = rows
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        for row in self.rows:
            self.get_packet_builder() \
                .add_int16(self.num_cols) \
                .add_row(row) \
                .write(write_file=write_file)


class NegotiateProtocolVersion(PostgresMessage):
    """
    NegotiateProtocolVersion (B)
    Byte1('v')
    Identifies the message as a protocol version negotiation message.

    Int32
    Length of message contents in bytes, including self.

    Int32
    Newest minor protocol version supported by the server for the major protocol version requested by the client.

    Int32
    Number of protocol options not recognized by the server.

    Then, for protocol option not recognized by the server, there is the following:

    String
    The option name. """

    def __init__(self, major_version, minor_version, option_not_recognized=None):
        self.identifier = PostgresBackendMessageIdentifier.NEGOTIATE_VERSION
        self.backend_capable = True
        self.frontend_capable = False
        self.major_version = major_version
        self.minor_version = minor_version
        self.option_not_recognized = option_not_recognized
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        packet_builder = self.get_packet_builder() \
            .add_int32(self.major_version) \
            .add_int32(self.minor_version)
        if self.option_not_recognized:
            packet_builder = packet_builder.add_string(self.option_not_recognized)

        packet_builder.write(write_file=write_file)


class ParseComplete(PostgresMessage):
    """
    ParseComplete (B)
    Byte1('1')
    Identifies the message as a Parse-complete indicator.

    Int32(4)
    Length of message contents in bytes, including self. """

    def __init__(self):
        self.identifier = PostgresBackendMessageIdentifier.PARSE_COMPLETE
        self.backend_capable = True
        self.frontend_capable = False
        super().__init__()

    def send_internal(self, write_file: BinaryIO):
        self.get_packet_builder() \
            .write(write_file=write_file)


class Query(PostgresMessage):
    """
    Query (F)
    Byte1('Q')
    Identifies the message as a simple query.

    Int32
    Length of message contents in bytes, including self.

    String
    The query string itself. """

    sql: bytes
    length: int

    def __init__(self):
        self.sql = b''
        self.length = 0
        self.identifier = PostgresFrontendMessageIdentifier.QUERY
        self.backend_capable = False
        self.frontend_capable = True
        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        self.length = packet_reader.read_int32()
        self.sql = packet_reader.read_bytes(self.length - 4)
        return self

    def get_parsed_sql(self, encoding=None):
        if not encoding:
            encoding = 'utf-8'
        try:
            sql = self.sql.decode('utf-8')
        except Exception:
            raise Exception(f'SQL contains non {encoding} values: {self.sql}')
        # Remove null bytes from end of sql statement. This is important.
        sql = strip_null_byte(sql)
        sql = SqlStatementParser.clear_sql(sql)
        return sql


class Terminate(PostgresMessage):
    """
    Terminate (F)
    Byte1('X')
    Identifies the message as a termination.

    Int32(4)
    Length of message contents in bytes, including self. """

    def __init__(self):
        self.identifier = PostgresFrontendMessageIdentifier.TERMINATE
        self.backend_capable = False
        self.frontend_capable = True
        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        return self


class BaseFrontendMessage(PostgresMessage):
    def __init__(self):
        self.backend_capable = False
        self.frontend_capable = True
        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        self.length = packet_reader.read_int32()
        if (self.length - 4) > 0:
            self.response = packet_reader.read_bytes(self.length - 4)
        return self


class Parse(BaseFrontendMessage):
    """
    Parse (F)
    Byte1('P')
    Identifies the message as a Parse command.

    Int32
    Length of message contents in bytes, including self.

    String
    The name of the destination prepared statement (an empty string selects the unnamed prepared statement).

    String
    The query string to be parsed.

    Int16 The number of parameter data types specified (can be zero). Note that this is not an indication of the
    number of parameters that might appear in the query string, only the number that the frontend wants to pre-specify
    types for.

    Then, for each parameter, there is the following:

    Int32 Specifies the object ID of the parameter data type. Placing a zero here is equivalent to leaving the type
    unspecified."""

    def __init__(self):
        self.identifier = PostgresFrontendMessageIdentifier.PARSE
        self.name = None
        self.query = None
        self.num_params = None
        self.parameters = []
        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        self.length = packet_reader.read_int32()
        self.name = packet_reader.read_string()
        self.query = packet_reader.read_string()
        self.num_params = packet_reader.read_int16()
        for i in range(self.num_params):
            self.parameters.append(packet_reader.read_int32())
        return self


class Bind(BaseFrontendMessage):
    """
    Bind (F)
    Byte1('B')
    Identifies the message as a Bind command.

    Int32
    Length of message contents in bytes, including self.

    String
    The name of the destination portal (an empty string selects the unnamed portal).

    String
    The name of the source prepared statement (an empty string selects the unnamed prepared statement).

    Int16 The number of parameter format codes that follow (denoted C below). This can be zero to indicate that there
    are no parameters or that the parameters all use the default format (text); or one, in which case the specified
    format code is applied to all parameters; or it can equal the actual number of parameters.

    Int16[C]
    The parameter format codes. Each must presently be zero (text) or one (binary).

    Int16 The number of parameter values that follow (possibly zero). This must match the number of parameters needed
    by the query.

    Next, the following pair of fields appear for each parameter:

    Int32 The length of the parameter value, in bytes (this count does not include itself). Can be zero. As a special
    case, -1 indicates a NULL parameter value. No value bytes follow in the NULL case.

    Byten
    The value of the parameter, in the format indicated by the associated format code. n is the above length.

    After the last parameter, the following fields appear:

    Int16 The number of result-column format codes that follow (denoted R below). This can be zero to indicate that
    there are no result columns or that the result columns should all use the default format (text); or one,
    in which case the specified format code is applied to all result columns (if any); or it can equal the actual
    number of result columns of the query.

    Int16[R]
    The result-column format codes. Each must presently be zero (text) or one (binary). """

    def __init__(self):
        self.identifier = PostgresFrontendMessageIdentifier.BIND
        self.length = None
        self.name = None
        self.statement_name = None
        self.format_codes = []  # 0=text 1=binary
        self.parameters = []
        self.result_format_codes = []

        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        self.length = packet_reader.read_int32()
        self.name = packet_reader.read_string()
        self.statement_name = packet_reader.read_string()
        num_format_codes = packet_reader.read_int16()
        for _ in range(num_format_codes):
            self.format_codes.append(packet_reader.read_int16())
        num_parameters = packet_reader.read_int16()
        for _ in range(num_parameters):
            param_length = packet_reader.read_int32()
            if param_length == -1:
                self.parameters.append(None)
            else:
                self.parameters.append(packet_reader.read_bytes(param_length))
        num_result_format_codes = packet_reader.read_int16()
        for _ in range(num_result_format_codes):
            self.result_format_codes.append(packet_reader.read_int16())
        return self


class Execute(BaseFrontendMessage):
    """
    Execute (F)
    Byte1('E')
    Identifies the message as an Execute command.

    Int32
    Length of message contents in bytes, including self.

    String
    The name of the portal to execute (an empty string selects the unnamed portal).

    Int32 Maximum number of rows to return, if portal contains a query that returns rows (ignored otherwise). Zero
    denotes “no limit”."""

    def __init__(self):
        self.identifier = PostgresFrontendMessageIdentifier.EXECUTE
        self.length = None
        self.name = None
        self.max_rows_ret = None
        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        self.length = packet_reader.read_int32()
        self.name = packet_reader.read_string()
        self.max_rows_ret = packet_reader.read_int32()
        return self


class Sync(BaseFrontendMessage):
    """
    Sync (F)
    Byte1('S')
    Identifies the message as a Sync command.

    Int32(4)
    Length of message contents in bytes, including self. """

    def __init__(self):
        self.identifier = PostgresFrontendMessageIdentifier.SYNC
        super().__init__()


class Describe(BaseFrontendMessage):
    """
    Describe (F)
    Byte1('D')
    Identifies the message as a Describe command.

    Int32
    Length of message contents in bytes, including self.

    Byte1
    'S' to describe a prepared statement; or 'P' to describe a portal.

    String The name of the prepared statement or portal to describe (an empty string selects the unnamed prepared
    statement or portal)."""

    def __init__(self):
        self.identifier = PostgresFrontendMessageIdentifier.DESCRIBE
        self.length = None
        self.describe_type = None
        self.name = None
        super().__init__()

    def read(self, packet_reader: PostgresPacketReader):
        self.length = packet_reader.read_int32()
        self.describe_type = packet_reader.read_byte()
        self.name = packet_reader.read_string()
        return self


IMPLEMENTED_BACKEND_POSTGRES_MESSAGE_CLASSES = [
    NoticeResponse, AuthenticationOk, AuthenticationClearTextPassword, ReadyForQuery, CommandComplete, Error,
    RowDescriptions, DataRow, NegotiateProtocolVersion, ParameterStatus, ParseComplete, BindComplete,
    ParameterDescription
]
IMPLEMENTED_FRONTEND_POSTGRES_MESSAGE_CLASSES = [
    Query, Terminate, Parse, Bind, Execute, Sync, Describe
]
FE_MESSAGE_MAP: Dict[PostgresFrontendMessageIdentifier, Type[PostgresMessage]] = {
    PostgresFrontendMessageIdentifier.QUERY: Query,
    PostgresFrontendMessageIdentifier.TERMINATE: Terminate,
    PostgresFrontendMessageIdentifier.PARSE: Parse,
    PostgresFrontendMessageIdentifier.BIND: Bind,
    PostgresFrontendMessageIdentifier.EXECUTE: Execute,
    PostgresFrontendMessageIdentifier.SYNC: Sync,
    PostgresFrontendMessageIdentifier.DESCRIBE: Describe
}
SUPPORTED_AUTH_TYPES = [PostgresAuthType.PASSWORD]

# Below Lies Unimplemented Messages

"""
'''
AuthenticationKerberosV5 (B)
Byte1('R')
Identifies the message as an authentication request.

Int32(8)
Length of message contents in bytes, including self.

Int32(2)
Specifies that Kerberos V5 authentication is required. '''

'''
AuthenticationMD5Password (B)
Byte1('R')
Identifies the message as an authentication request.

Int32(12)
Length of message contents in bytes, including self.

Int32(5)
Specifies that an MD5-encrypted password is required.

Byte4
The salt to use when encrypting the password. '''

'''
AuthenticationSCMCredential (B)
Byte1('R')
Identifies the message as an authentication request.

Int32(8)
Length of message contents in bytes, including self.

Int32(6)
Specifies that an SCM credentials message is required. '''

'''
AuthenticationGSS (B)
Byte1('R')
Identifies the message as an authentication request.

Int32(8)
Length of message contents in bytes, including self.

Int32(7)
Specifies that GSSAPI authentication is required. '''

'''
AuthenticationGSSContinue (B)
Byte1('R')
Identifies the message as an authentication request.

Int32
Length of message contents in bytes, including self.

Int32(8)
Specifies that this message contains GSSAPI or SSPI data.

Byten
GSSAPI or SSPI authentication data. '''

'''
AuthenticationSSPI (B)
Byte1('R')
Identifies the message as an authentication request.

Int32(8)
Length of message contents in bytes, including self.

Int32(9)
Specifies that SSPI authentication is required. '''

'''
AuthenticationSASL (B)
Byte1('R')
Identifies the message as an authentication request.

Int32
Length of message contents in bytes, including self.

Int32(10)
Specifies that SASL authentication is required.

The message body is a list of SASL authentication mechanisms, in the server's order of preference. A zero byte is
required as terminator after the last authentication mechanism name. For each mechanism, there is the following:

String
Name of a SASL authentication mechanism. '''

'''
AuthenticationSASLContinue (B)
Byte1('R')
Identifies the message as an authentication request.

Int32
Length of message contents in bytes, including self.

Int32(11)
Specifies that this message contains a SASL challenge.

Byten
SASL data, specific to the SASL mechanism being used. '''

'''
AuthenticationSASLFinal (B)
Byte1('R')
Identifies the message as an authentication request.

Int32
Length of message contents in bytes, including self.

Int32(12)
Specifies that SASL authentication has completed.

Byten
SASL outcome "additional data", specific to the SASL mechanism being used. '''

'''BackendKeyData (B) Byte1('K') Identifies the message as cancellation key data. The frontend must save these values
if it wishes to be able to issue CancelRequest messages later.

Int32(12)
Length of message contents in bytes, including self.

Int32
The process ID of this backend.

Int32
The secret key of this backend. '''

'''
CancelRequest (F)
Int32(16)
Length of message contents in bytes, including self.

Int32(80877102) The cancel request code. The value is chosen to contain 1234 in the most significant 16 bits,
and 5678 in the least significant 16 bits. (To avoid confusion, this code must not be the same as any protocol
version number.)

Int32
The process ID of the target backend.

Int32
The secret key for the target backend. '''

'''
Close (F)
Byte1('C')
Identifies the message as a Close command.

Int32
Length of message contents in bytes, including self.

Byte1
'S' to close a prepared statement; or 'P' to close a portal.

String The name of the prepared statement or portal to close (an empty string selects the unnamed prepared statement
or portal).'''

'''
CloseComplete (B)
Byte1('3')
Identifies the message as a Close-complete indicator.

Int32(4)
Length of message contents in bytes, including self. '''

'''
CommandComplete (B)
Byte1('C')
Identifies the message as a command-completed response.

Int32
Length of message contents in bytes, including self.

String
The command tag. This is usually a single word that identifies which SQL command was completed.

For an INSERT command, the tag is INSERT oid rows, where rows is the number of rows inserted. oid used to be the
object ID of the inserted row if rows was 1 and the target table had OIDs, but OIDs system columns are not supported
anymore; therefore oid is always 0.

For a DELETE command, the tag is DELETE rows where rows is the number of rows deleted.

For an UPDATE command, the tag is UPDATE rows where rows is the number of rows updated.

For a SELECT or CREATE TABLE AS command, the tag is SELECT rows where rows is the number of rows retrieved.

For a MOVE command, the tag is MOVE rows where rows is the number of rows the cursor's position has been changed by.

For a FETCH command, the tag is FETCH rows where rows is the number of rows that have been retrieved from the cursor.

For a COPY command, the tag is COPY rows where rows is the number of rows copied. (Note: the row count appears only
in PostgreSQL 8.2 and later.)'''

'''
CopyData (F & B)
Byte1('d')
Identifies the message as COPY data.

Int32
Length of message contents in bytes, including self.

Byten Data that forms part of a COPY data stream. Messages sent from the backend will always correspond to single
data rows, but messages sent by frontends might divide the data stream arbitrarily.'''

'''
CopyDone (F & B)
Byte1('c')
Identifies the message as a COPY-complete indicator.

Int32(4)
Length of message contents in bytes, including self. '''

'''
CopyFail (F)
Byte1('f')
Identifies the message as a COPY-failure indicator.

Int32
Length of message contents in bytes, including self.

String
An error message to report as the cause of failure. '''

'''CopyInResponse (B) Byte1('G') Identifies the message as a Start Copy In response. The frontend must now send
copy-in data (if not prepared to do so, send a CopyFail message).

Int32
Length of message contents in bytes, including self.

Int8 0 indicates the overall COPY format is textual (rows separated by newlines, columns separated by separator
characters, etc.). 1 indicates the overall copy format is binary (similar to DataRow format). See COPY for more
information.

Int16
The number of columns in the data to be copied (denoted N below).

Int16[N] The format codes to be used for each column. Each must presently be zero (text) or one (binary). All must be
zero if the overall copy format is textual.'''

'''
CopyOutResponse (B)
Byte1('H')
Identifies the message as a Start Copy Out response. This message will be followed by copy-out data.

Int32
Length of message contents in bytes, including self.

Int8 0 indicates the overall COPY format is textual (rows separated by newlines, columns separated by separator
characters, etc.). 1 indicates the overall copy format is binary (similar to DataRow format). See COPY for more
information.

Int16
The number of columns in the data to be copied (denoted N below).

Int16[N] The format codes to be used for each column. Each must presently be zero (text) or one (binary). All must be
zero if the overall copy format is textual.'''

'''
CopyBothResponse (B)
Byte1('W')
Identifies the message as a Start Copy Both response. This message is used only for Streaming Replication.

Int32
Length of message contents in bytes, including self.

Int8 0 indicates the overall COPY format is textual (rows separated by newlines, columns separated by separator
characters, etc.). 1 indicates the overall copy format is binary (similar to DataRow format). See COPY for more
information.

Int16
The number of columns in the data to be copied (denoted N below).

Int16[N] The format codes to be used for each column. Each must presently be zero (text) or one (binary). All must be
zero if the overall copy format is textual.'''

'''
Describe (F)
Byte1('D')
Identifies the message as a Describe command.

Int32
Length of message contents in bytes, including self.

Byte1
'S' to describe a prepared statement; or 'P' to describe a portal.

String The name of the prepared statement or portal to describe (an empty string selects the unnamed prepared
statement or portal).'''

'''
EmptyQueryResponse (B)
Byte1('I')
Identifies the message as a response to an empty query string. (This substitutes for CommandComplete.)

Int32(4)
Length of message contents in bytes, including self. '''

'''
Execute (F)
Byte1('E')
Identifies the message as an Execute command.

Int32
Length of message contents in bytes, including self.

String
The name of the portal to execute (an empty string selects the unnamed portal).

Int32 Maximum number of rows to return, if portal contains a query that returns rows (ignored otherwise). Zero
denotes “no limit”.'''

'''
Flush (F)
Byte1('H')
Identifies the message as a Flush command.

Int32(4)
Length of message contents in bytes, including self. '''

'''
FunctionCall (F)
Byte1('F')
Identifies the message as a function call.

Int32
Length of message contents in bytes, including self.

Int32
Specifies the object ID of the function to call.

Int16 The number of argument format codes that follow (denoted C below). This can be zero to indicate that there are
no arguments or that the arguments all use the default format (text); or one, in which case the specified format code
is applied to all arguments; or it can equal the actual number of arguments.

Int16[C]
The argument format codes. Each must presently be zero (text) or one (binary).

Int16
Specifies the number of arguments being supplied to the function.

Next, the following pair of fields appear for each argument:

Int32 The length of the argument value, in bytes (this count does not include itself). Can be zero. As a special
case, -1 indicates a NULL argument value. No value bytes follow in the NULL case.

Byten
The value of the argument, in the format indicated by the associated format code. n is the above length.

After the last argument, the following field appears:

Int16
The format code for the function result. Must presently be zero (text) or one (binary). '''

'''
FunctionCallResponse (B)
Byte1('V')
Identifies the message as a function call result.

Int32
Length of message contents in bytes, including self.

Int32 The length of the function result value, in bytes (this count does not include itself). Can be zero. As a
special case, -1 indicates a NULL function result. No value bytes follow in the NULL case.

Byten
The value of the function result, in the format indicated by the associated format code. n is the above length. '''

'''
GSSENCRequest (F)
Int32(8)
Length of message contents in bytes, including self.

Int32(80877104) The GSSAPI Encryption request code. The value is chosen to contain 1234 in the most significant 16
bits, and 5680 in the least significant 16 bits. (To avoid confusion, this code must not be the same as any protocol
version number.)'''

'''GSSResponse (F) Byte1('p') Identifies the message as a GSSAPI or SSPI response. Note that this is also used for
SASL and password response messages. The exact message type can be deduced from the context.

Int32
Length of message contents in bytes, including self.

Byten
GSSAPI/SSPI specific message data. '''

'''
NoData (B)
Byte1('n')
Identifies the message as a no-data indicator.

Int32(4)
Length of message contents in bytes, including self. '''

'''
NotificationResponse (B)
Byte1('A')
Identifies the message as a notification response.

Int32
Length of message contents in bytes, including self.

Int32
The process ID of the notifying backend process.

String
The name of the channel that the notify has been raised on.

String
The “payload” string passed from the notifying process. '''

'''
Parse (F)
Byte1('P')
Identifies the message as a Parse command.

Int32
Length of message contents in bytes, including self.

String
The name of the destination prepared statement (an empty string selects the unnamed prepared statement).

String
The query string to be parsed.

Int16 The number of parameter data types specified (can be zero). Note that this is not an indication of the number
of parameters that might appear in the query string, only the number that the frontend wants to prespecify types for.

Then, for each parameter, there is the following:

Int32 Specifies the object ID of the parameter data type. Placing a zero here is equivalent to leaving the type
unspecified.'''

'''PasswordMessage (F) Byte1('p') Identifies the message as a password response. Note that this is also used for
GSSAPI, SSPI and SASL response messages. The exact message type can be deduced from the context.

Int32
Length of message contents in bytes, including self.

String
The password (encrypted, if requested). '''

'''PortalSuspended (B) Byte1('s') Identifies the message as a portal-suspended indicator. Note this only appears if
an Execute message's row-count limit was reached.

Int32(4)
Length of message contents in bytes, including self. '''

'''SASLInitialResponse (F) Byte1('p') Identifies the message as an initial SASL response. Note that this is also used
for GSSAPI, SSPI and password response messages. The exact message type is deduced from the context.

Int32
Length of message contents in bytes, including self.

String
Name of the SASL authentication mechanism that the client selected.

Int32
Length of SASL mechanism specific "Initial Client Response" that follows, or -1 if there is no Initial Response.

Byten
SASL mechanism specific "Initial Response". '''

'''SASLResponse (F) Byte1('p') Identifies the message as a SASL response. Note that this is also used for GSSAPI,
SSPI and password response messages. The exact message type can be deduced from the context.

Int32
Length of message contents in bytes, including self.

Byten
SASL mechanism specific message data. '''

'''
StartupMessage (F)
Int32
Length of message contents in bytes, including self.

Int32(196608) The protocol version number. The most significant 16 bits are the major version number (3 for the
protocol described here). The least significant 16 bits are the minor version number (0 for the protocol described
here).

The protocol version number is followed by one or more pairs of parameter name and value strings. A zero byte is
required as a terminator after the last name/value pair. Parameters can appear in any order. user is required,
others are optional. Each parameter is specified as:

String
The parameter name. Currently recognized names are:

user
The database user name to connect as. Required; there is no default.

database
The database to connect to. Defaults to the user name.

options Command-line arguments for the backend. (This is deprecated in favor of setting individual run-time
parameters.) Spaces within this string are considered to separate arguments, unless escaped with a backslash
write to represent a literal backslash.

replication Used to connect in streaming replication mode, where a small set of replication commands can be issued
instead of SQL statements. Value can be true, false, or database, and the default is false. See Section 55.4 for
details.

In addition to the above, other parameters may be listed. Parameter names beginning with _pq_. are reserved for use
as protocol extensions, while others are treated as run-time parameters to be set at backend start time. Such
settings will be applied during backend start (after parsing the command-line arguments if any) and will act as
session defaults.

String
The parameter value. '''
"""



import unittest

from mindsdb.integrations.handlers.maxdb_handler.maxdb_handler import MaxDBHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class SurrealdbHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "127.0.0.1",
                "port": "7210",
                "user": "MAXDB",
                "password": "password",
                "database": "MAXDB",
                "jdbc_location": "/path/to/jdbc/sapdbc.jar"
            }
        }
        cls.handler = MaxDBHandler('test_maxdb_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_create_table(self):
        res = self.handler.native_query("CREATE TABLE TEST_TABLE (id INT PRIMARY KEY,name VARCHAR(50))")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_2_insert(self):
        res = self.handler.native_query("INSERT INTO TEST_TABLE (id, name) VALUES (1, 'MARSID')")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_3_select_query(self):
        query = "SELECT * FROM TEST_TABLE"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_get_columns(self):
        columns = self.handler.get_columns('TEST_TABLE')
        assert columns.type is not RESPONSE_TYPE.ERROR

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_6_drop_table(self):
        res = self.handler.native_query("DROP TABLE TEST_TABLE")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_7_disconnect(self):
        assert self.handler.disconnect() is None


if __name__ == '__main__':
    unittest.main()



__title__ = "MindsDB Lindorm handler"
__package_name__ = "mindsdb_lindorm_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for Lindorm"
__author__ = "Someshfengde"
__github__ = "https://github.com/someshfengde"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - mindsdb"



import os

import re
from twilio.rest import Client
import pandas as pd
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities.config import Config
from mindsdb.utilities import log
from mindsdb.integrations.utilities.date_utils import parse_local_date
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions, project_dataframe, filter_dataframe

from mindsdb_sql.parser import ast

logger = log.getLogger(__name__)


class PhoneNumbersTable(APITable):

    def select(self, query: ast.Select) -> Response:

        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:

            if op == 'or':
                raise NotImplementedError('OR is not supported')
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params['limit'] = query.limit.value

        result = self.handler.list_phone_numbers(params, df=True)

        # filter targets
        result = filter_dataframe(result, filters)

        # project targets
        result = project_dataframe(result, query.targets, self.get_columns())

        return result

    def get_columns(self):
        return [
            'sid',
            'date_created',
            'date_updated',
            'phone_number',
            'friendly_name',
            'account_sid',
            'capabilities',
            'number_status',
            'api_version',
            'voice_url',
            'sms_url',
            'uri'
        ]


class MessagesTable(APITable):

    def select(self, query: ast.Select) -> Response:

        conditions = extract_comparison_conditions(query.where)

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:

            if op == 'or':
                raise NotImplementedError('OR is not supported')
            if arg1 == 'sent_at' and arg2 is not None:

                date = parse_local_date(arg2)

                if op == '>':
                    params['date_sent_after'] = date
                elif op == '<':
                    params['date_sent_before'] = date
                else:
                    raise NotImplementedError

                # also add to post query filter because date_sent_after=date1 will include date1
                filters.append([op, arg1, arg2])

            elif arg1 == 'sid':
                if op == '=':
                    params['sid'] = arg2
                # TODO: implement IN
                else:
                    NotImplementedError('Only  "from_number=" is implemented')
            elif arg1 == 'from_number':
                if op == '=':
                    params['from_number'] = arg2
                # TODO: implement IN
                else:
                    NotImplementedError('Only  "from_number=" is implemented')

            elif arg1 == 'to_number':
                if op == '=':
                    params['to_number'] = arg2
                # TODO: implement IN
                else:
                    NotImplementedError('Only  "to_number=" is implemented')

            else:
                filters.append([op, arg1, arg2])

        result = self.handler.fetch_messages(params, df=True)

        # filter targets
        result = filter_dataframe(result, filters)

        if query.limit is not None:
            result = result[:int(query.limit.value)]

        # project targets
        result = project_dataframe(result, query.targets, self.get_columns())

        return result

    def get_columns(self):
        return [
            'sid',
            'from_number',
            'to_number',
            'body',
            'direction',
            'msg_status',
            'sent_at',  # datetime.strptime(str(msg.date_sent), '%Y-%m-%d %H:%M:%S%z'),
            'account_sid',
            'price',
            'price_unit',
            'api_version',
            'uri'
        ]

    def insert(self, query: ast.Insert):
        # https://docs.tweepy.org/en/stable/client.html#tweepy.Client.create_tweet
        columns = [col.name for col in query.columns]

        ret = []

        insert_params = ["to_number", "from_number", "body", 'media_url']
        for row in query.values:
            params = dict(zip(columns, row))

            # split long text over 1500 symbols
            max_text_len = 1500
            text = params['body']
            words = re.split('( )', text)
            messages = []

            text2 = ''
            pattern = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
            for word in words:
                # replace the links in word to string with the length as twitter short url (23)
                word2 = re.sub(pattern, '-' * 23, word)
                if len(text2) + len(word2) > max_text_len - 3 - 7:  # 3 is for ..., 7 is for (10/11)
                    messages.append(text2.strip())

                    text2 = ''
                text2 += word

            # the last message
            if text2.strip() != '':
                messages.append(text2.strip())

            len_messages = len(messages)

            for i, text in enumerate(messages):
                if i < len_messages - 1:
                    text += '...'
                else:
                    text += ' '

                if i >= 1:
                    text += f'({i + 1}/{len_messages})'
                    # only send image on first url
                    if 'media_url' in params:
                        del params['media_url']

                params['body'] = text
                params_to_send = {key: params[key] for key in insert_params if (key in params)}
                ret_row = self.handler.send_sms(params_to_send, ret_as_dict=True)
                ret_row['body'] = text
                ret.append(ret_row)

        return pd.DataFrame(ret)


class TwilioHandler(APIHandler):

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get('connection_data', {})

        self.connection_args = {}
        handler_config = Config().get('twilio_handler', {})
        for k in ['account_sid', 'auth_token']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'TWILIO_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'TWILIO_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.client = None
        self.is_connected = False

        messages = MessagesTable(self)
        phone_numbers = PhoneNumbersTable(self)
        self._register_table('messages', messages)
        self._register_table('phone_numbers', phone_numbers)

    def connect(self):
        """Authenticate with the Twilio API using the account_sid and auth_token provided in the constructor."""
        if self.is_connected is True:
            return self.client

        self.client = Client(
            self.connection_args['account_sid'],
            self.connection_args['auth_token']
        )

        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        '''It evaluates if the connection with Twilio API is alive and healthy.'''
        response = StatusResponse(False)

        try:
            self.connect()
            # Maybe make a harmless API request to verify connection, but be mindful of rate limits and costs
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Twilio api: {e}. '
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def parse_native_query(self, query_string: str):
        """Parses the native query string of format method(arg1=val1, arg2=val2, ...) and returns the method name and arguments."""

        # Adjust regex to account for the possibility of no arguments inside the parenthesis
        match = re.match(r'(\w+)\(([^)]*)\)', query_string)
        if not match:
            raise ValueError(f"Invalid query format: {query_string}")

        method_name = match.group(1)
        arg_string = match.group(2)

        # Extract individual arguments
        args = {}
        if arg_string:  # Check if there are any arguments
            for arg in arg_string.split(','):
                arg = arg.strip()
                key, value = arg.split('=')
                args[key.strip()] = value.strip()

        return method_name, args

    def native_query(self, query_string: str = None):
        '''It parses any native statement string and acts upon it (for example, raw syntax commands).'''

        method_name, params = self.parse_native_query(query_string)
        if method_name == 'send_sms':
            response = self.send_sms(params)
        elif method_name == 'fetch_messages':
            response = self.fetch_messages(params)
        elif method_name == 'list_phone_numbers':
            response = self.list_phone_numbers(params)
        else:
            raise ValueError(f"Method '{method_name}' not supported by TwilioHandler")

        return response

    def send_sms(self, params, ret_as_dict=False):
        message = self.client.messages.create(
            to=params.get("to_number"),
            from_=params.get('from_number'),
            body=params.get("body"),
            media_url=params.get("media_url")
        )

        if ret_as_dict is True:
            return {'sid': message.sid, 'status': message.status}
        return Response(
            RESPONSE_TYPE.MESSAGE,
            sid=message.sid,
            status=message.status
        )

    def fetch_messages(self, params, df=False):
        limit = int(params.get('limit', 1000))
        sid = params.get('sid', None)
        # Convert date strings to datetime objects if provided
        date_sent_after = params.get('date_sent_after', None)
        date_sent_before = params.get('date_sent_before', None)
        # Extract 'from_' and 'body' search criteria from params
        from_number = params.get('from_number', None)
        to_number = params.get('to_number', None)
        args = {
            'limit': limit,
            'date_sent_after': date_sent_after,
            'date_sent_before': date_sent_before,
            'from_': from_number,
            'to': to_number
        }

        args = {arg: val for arg, val in args.items() if val is not None}
        if sid:
            messages = [self.client.messages(sid).fetch()]
        else:
            messages = self.client.messages.list(**args)

        # Extract all possible properties for each message
        data = []
        for msg in messages:
            msg_data = {
                'sid': msg.sid,
                'to_number': msg.to,
                'from_number': msg.from_,
                'body': msg.body,
                'direction': msg.direction,
                'msg_status': msg.status,
                'sent_at': msg.date_created.replace(tzinfo=None),
                'account_sid': msg.account_sid,
                'price': msg.price,
                'price_unit': msg.price_unit,
                'api_version': msg.api_version,
                'uri': msg.uri,
                # 'media_url': [media.uri for media in msg.media.list()]
                # ... Add other properties as needed
            }
            data.append(msg_data)

        if df is True:
            return pd.DataFrame(data)
        return Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(data))

    def list_phone_numbers(self, params, df=False):
        limit = int(params.get('limit', 100))
        args = {
            'limit': limit
        }
        args = {arg: val for arg, val in args.items() if val is not None}
        phone_numbers = self.client.incoming_phone_numbers.list(**args)

        # Extract properties for each phone number
        data = []
        for number in phone_numbers:
            num_data = {
                'sid': number.sid,
                'date_created': number.date_created,
                'date_updated': number.date_updated,
                'phone_number': number.phone_number,
                'friendly_name': number.friendly_name,
                'account_sid': number.account_sid,
                'capabilities': number.capabilities,
                'number_status': number.status,
                'api_version': number.api_version,
                'voice_url': number.voice_url,
                'sms_url': number.sms_url,
                'uri': number.uri,
                # ... Add other properties as needed
            }
            data.append(num_data)

        if df is True:
            return pd.DataFrame(data)
        return Response(RESPONSE_TYPE.TABLE, data_frame=pd.DataFrame(data))



import pandas as pd
from typing import Dict

from openbb import obb

from mindsdb.integrations.handlers.openbb_handler.openbb_tables import OpenBBtable, create_table_class
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
from mindsdb_sql import parse_sql

logger = log.getLogger(__name__)


class OpenBBHandler(APIHandler):
    """A class for handling connections and interactions with the OpenBB Platform.

    Attributes:
        PAT (str): OpenBB's personal access token. Sign up here: https://my.openbb.co
        is_connected (bool): Whether or not the user is connected to their OpenBB account.

    """

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        self.PAT = None

        args = kwargs.get("connection_data", {})
        if "PAT" in args:
            self.PAT = args["PAT"]

        self.is_connected = False

        self.obb = obb
        from openbb_provider import standard_models

        self._register_table("openbb_fetcher", OpenBBtable(self))
        stock_table_class = create_table_class(
            params_metadata=standard_models.stock_historical.StockHistoricalQueryParams.model_json_schema(),
            response_metadata=standard_models.stock_historical.StockHistoricalData.model_json_schema(),
            obb_function=self.obb.stocks.load
        )
        self._register_table("stock_historical", stock_table_class(self))

        crypto_table_class = create_table_class(
            params_metadata=standard_models.crypto_historical.CryptoHistoricalQueryParams.model_json_schema(),
            response_metadata=standard_models.crypto_historical.CryptoHistoricalData.model_json_schema(),
            obb_function=self.obb.crypto.load
        )
        self._register_table("crypto_historical", crypto_table_class(self))

    def connect(self) -> bool:
        """Connects with OpenBB account through personal access token (PAT).

        Returns none.
        """
        self.is_connected = False
        obb.account.login(pat=self.PAT)

        # Check if PAT utilized is valid
        # if obb.user.profile.active:
        self.is_connected = True
        return True

    def check_connection(self) -> StatusResponse:
        """Checks connection to OpenBB accounting by checking the validity of the PAT.

        Returns StatusResponse indicating whether or not the handler is connected.
        """

        response = StatusResponse(False)

        try:
            if self.connect():
                response.success = True

        except Exception as e:
            logger.error(f"Error connecting to OpenBB Platform: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def _process_cols_names(self, cols: list) -> list:
        new_cols = []
        for element in cols:
            # If the element is a tuple, we want to merge the elements together
            if isinstance(element, tuple):
                # If there's more than one element we want to merge them together
                if len(element) > 1:
                    # Prevents the case where there's a multi column index and the index is a date
                    # in that instance we will have ('date', '') and this avoids having a column named 'date_'
                    new_element = "_".join(map(str, element)).rstrip("_")
                    new_cols.append(new_element)
                else:
                    new_cols.append(element[0])
            else:
                new_cols.append(element)
        return new_cols

    def _openbb_fetcher(self, params: Dict = None) -> pd.DataFrame:
        """Gets aggregate trade data for a symbol based on given parameters

        Returns results as a pandas DataFrame.

        Args:
            params (Dict): Trade data params (symbol, interval, limit, start_time, end_time)
        """
        self.connect()

        try:
            if params is None:
                logger.error("At least cmd needs to be added!")
                raise Exception("At least cmd needs to be added!")

            # Get the OpenBB command to get the data from
            cmd = params.pop("cmd")

            # Ensure that the cmd provided is a valid OpenBB command
            available_cmds = [f"obb{cmd}" for cmd in list(obb.coverage.commands.keys())]
            if cmd not in available_cmds:
                logger.error(f"The command provided is not supported by OpenBB! Choose one of the following: {', '.join(available_cmds)}")
                raise Exception(f"The command provided is not supported by OpenBB! Choose one of the following: {', '.join(available_cmds)}")

            args = ""
            # If there are parameters create arguments as a string
            if params:
                for arg, val in params.items():
                    args += f"{arg}={val},"

                # Remove the additional ',' added at the end
                if args:
                    args = args[:-1]

            # Recreate the OpenBB command with the arguments
            openbb_cmd = f"{cmd}({args})"

            # Execute the OpenBB command and return the OBBject
            openbb_object = eval(openbb_cmd)

            # Transform the OBBject into a pandas DataFrame
            data = openbb_object.to_df()

            # Check if index is a datetime, if it is we want that as a column
            if isinstance(data.index, pd.DatetimeIndex):
                data.reset_index(inplace=True)

            # Process column names
            data.columns = self._process_cols_names(data.columns)

        except Exception as e:
            logger.error(f"Error accessing data from OpenBB: {e}!")
            raise Exception(f"Error accessing data from OpenBB: {e}!")

        return data

    def native_query(self, query: str = None) -> Response:
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)

    def call_openbb_api(
        self, method_name: str = None, params: Dict = None
    ) -> pd.DataFrame:
        """Calls the OpenBB Platform method with the given params.

        Returns results as a pandas DataFrame.

        Args:
            method_name (str): Method name to call (e.g. klines)
            params (Dict): Params to pass to the API call
        """
        if method_name == "openbb_fetcher":
            return self._openbb_fetcher(params)
        raise NotImplementedError(
            "Method name {} not supported by OpenBB Platform Handler".format(
                method_name
            )
        )



import paypalrestsdk
import pandas as pd
from typing import Text, List, Dict

from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor


class PaymentsTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls PayPal Payments data.
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            PayPal Payments matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'payments',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        payments_df = pd.json_normalize(self.get_payments(count=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            payments_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        payments_df = select_statement_executor.execute_query()

        return payments_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_payments(count=1)).columns.tolist()

    def get_payments(self, **kwargs) -> List[Dict]:
        connection = self.handler.connect()
        payments = paypalrestsdk.Payment.all(kwargs, api=connection)
        return [payment.to_dict() for payment in payments['payments']]


class InvoicesTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:
        select_statement_parser = SELECTQueryParser(
            query,
            'invoices',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        invoices_df = pd.json_normalize(self.get_invoices(count=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            invoices_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        invoices_df = select_statement_executor.execute_query()

        return invoices_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_invoices(count=1)).columns.tolist()

    def get_invoices(self, **kwargs) -> List[Dict]:
        connection = self.handler.connect()
        invoices = paypalrestsdk.Invoice.all(kwargs, api=connection)
        return [invoice.to_dict() for invoice in invoices['invoices']]


class SubscriptionsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        select_statement_parser = SELECTQueryParser(
            query,
            'subscriptions',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        subscriptions_df = pd.json_normalize(self.get_subscriptions(count=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            subscriptions_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        subscriptions_df = select_statement_executor.execute_query()
        return subscriptions_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_subscriptions(count = 1)).columns.tolist()

    def get_subscriptions(self, **kwargs) -> List[Dict]:
        connection = self.handler.connect()
        subscriptions = paypalrestsdk.BillingPlan.all(kwargs, api=connection)
        return [subscription.to_dict() for subscription in subscriptions['plans']]


class OrdersTable(APITable):
    """The PayPal Orders Table implementation"""
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls PayPal Orders data.
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            PayPal Orders matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'orders',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {"ids": []}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'ids':
                if op == '=':
                    search_params["ids"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for 'ids' column")
            elif arg1 in ['state', 'amount', 'create_time', 'update_time', 'links', 'pending_reason', 'parent_payment']:
                subset_where_conditions.append([op, arg1, arg2])

        if search_params == {}:
            raise NotImplementedError("id column is required for this table")

        orders_df = pd.json_normalize(self.get_orders(search_params))
        self.clean_selected_columns(selected_columns)
        select_statement_executor = SELECTQueryExecutor(
            orders_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions
        )
        orders_df = select_statement_executor.execute_query()
        return orders_df

    @staticmethod
    def clean_selected_columns(selected_cols) -> None:
        if "ids" in selected_cols:
            selected_cols.remove("ids")
            selected_cols.append("id")

    def get_columns(self) -> List[Text]:
        return ["id", "state", "amount", "create_time", "update_time", "links", "pending_reason", "parent_payment"]

    def get_orders(self, kwargs) -> List[Dict]:
        connection = self.handler.connect()
        orders = []
        for value in kwargs["ids"]:
            try:
                order = paypalrestsdk.Order.find(value, api=connection)
            except paypalrestsdk.exceptions.ResourceNotFound:
                continue
            orders.append(order.to_dict())
        return orders


class PayoutsTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls PayPal payouts data.
        Parameters
        ----------
        query : ast.Select
            Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            PayPal payouts matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'payouts',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        conditions = extract_comparison_conditions(query.where)

        payout_batch_id = ""

        for a_where in conditions:
            if a_where[1] == "payout_batch_id":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for state")

                payout_batch_id = a_where[2]

        payouts_data = self.get_payout(payout_batch_id)  # Get the data
        payouts_df = pd.DataFrame(payouts_data)  # Create a DataFrame

        select_statement_executor = SELECTQueryExecutor(
            payouts_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        payouts_df = select_statement_executor.execute_query()

        return payouts_df

    def get_columns(self) -> List[Text]:
        return [
            "payout_batch_id",
            "batch_status",
            "time_created",
            "time_completed",
            "sender_batch_id",
            "email_subject",
            "email_message",
            "funding_source",
            "amount_currency",
            "amount_value",
            "fees_currency",
            "fees_value",
        ]

    def get_payout(self, payout_batch_id:str) -> List[Dict]:
        connection = self.handler.connect()
        endpoint = f"v1/payments/payouts/{payout_batch_id}"
        payout = connection.get(endpoint)

        payout_data = {
            "payout_batch_id": payout['batch_header']['payout_batch_id'],
            "batch_status": payout['batch_header']['batch_status'],
            "time_created": payout['batch_header']['time_created'],
            "time_completed": payout['batch_header']['time_completed'],
            "sender_batch_id": payout['batch_header']['sender_batch_header']['sender_batch_id'],
            "email_subject": payout['batch_header']['sender_batch_header']['email_subject'],
            "email_message": payout['batch_header']['sender_batch_header']['email_message'],
            "funding_source": payout['batch_header']['funding_source'],
            "amount_currency": payout['batch_header']['amount']['currency'],
            "amount_value": payout['batch_header']['amount']['value'],
            "fees_currency": payout['batch_header']['fees']['currency'],
            "fees_value": payout['batch_header']['fees']['value'],
        }

        return [payout_data]




import numpy as np
from numpy import dtype as np_dtype
import pandas as pd
from pandas.api import types as pd_types
from sqlalchemy.types import (
    Integer, Float, Text
)

from mindsdb_sql.parser.ast import Insert, Identifier, CreateTable, TableColumn, DropTables

from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow
from mindsdb.utilities import log
from mindsdb.utilities.profiler import profiler

logger = log.getLogger(__name__)


class DBHandlerException(Exception):
    pass


class IntegrationDataNode(DataNode):
    type = 'integration'

    def __init__(self, integration_name, ds_type, integration_controller):
        self.integration_name = integration_name
        self.ds_type = ds_type
        self.integration_controller = integration_controller
        self.integration_handler = self.integration_controller.get_handler(self.integration_name)

    def get_type(self):
        return self.type

    def get_tables(self):
        response = self.integration_handler.get_tables()
        if response.type == RESPONSE_TYPE.TABLE:
            result_dict = response.data_frame.to_dict(orient='records')
            result = []
            for row in result_dict:
                result.append(TablesRow.from_dict(row))
            return result
        else:
            raise Exception(f"Can't get tables: {response.error_message}")

    def has_table(self, tableName):
        return True

    def get_table_columns(self, tableName):
        return []

    def drop_table(self, name: Identifier, if_exists=False):
        drop_ast = DropTables(
            tables=[name],
            if_exists=if_exists
        )
        result = self.integration_handler.query(drop_ast)
        if result.type == RESPONSE_TYPE.ERROR:
            raise Exception(result.error_message)

    def create_table(self, table_name: Identifier, result_set, is_replace=False, is_create=False):
        # is_create - create table
        # is_replace - drop table if exists
        # is_create==False and is_replace==False: just insert

        table_columns_meta = {}
        table_columns = []
        for col in result_set.columns:
            # assume this is pandas type
            column_type = Text
            if isinstance(col.type, np_dtype):
                if pd_types.is_integer_dtype(col.type):
                    column_type = Integer
                elif pd_types.is_numeric_dtype(col.type):
                    column_type = Float

            table_columns.append(
                TableColumn(
                    name=col.alias,
                    type=column_type
                )
            )
            table_columns_meta[col.alias] = column_type

        if is_replace:
            # drop
            drop_ast = DropTables(
                tables=[table_name],
                if_exists=True
            )
            result = self.integration_handler.query(drop_ast)
            if result.type == RESPONSE_TYPE.ERROR:
                raise Exception(result.error_message)
            is_create = True

        if is_create:
            create_table_ast = CreateTable(
                name=table_name,
                columns=table_columns,
                is_replace=True
            )

            result = self.integration_handler.query(create_table_ast)
            if result.type == RESPONSE_TYPE.ERROR:
                raise Exception(result.error_message)

        insert_columns = [Identifier(parts=[x.alias]) for x in result_set.columns]
        formatted_data = []

        for rec in result_set.get_records():
            new_row = []
            for col in result_set.columns:
                value = rec[col.alias]
                column_type = table_columns_meta[col.alias]

                python_type = str
                if column_type == Integer:
                    python_type = int
                elif column_type == Float:
                    python_type = float

                try:
                    value = python_type(value) if value is not None else value
                except Exception:
                    pass
                new_row.append(value)
            formatted_data.append(new_row)

        if len(formatted_data) == 0:
            # not need to insert
            return

        insert_ast = Insert(
            table=table_name,
            columns=insert_columns,
            values=formatted_data
        )

        try:
            result = self.integration_handler.query(insert_ast)
        except Exception as e:
            msg = f'[{self.ds_type}/{self.integration_name}]: {str(e)}'
            raise DBHandlerException(msg) from e

        if result.type == RESPONSE_TYPE.ERROR:
            raise Exception(result.error_message)

    @profiler.profile()
    def query(self, query=None, native_query=None, session=None):
        try:
            if query is not None:
                result = self.integration_handler.query(query)
            else:
                # try to fetch native query
                result = self.integration_handler.native_query(native_query)
        except Exception as e:
            msg = str(e).strip()
            if msg == '':
                msg = e.__class__.__name__
            msg = f'[{self.ds_type}/{self.integration_name}]: {msg}'
            raise DBHandlerException(msg) from e

        if result.type == RESPONSE_TYPE.ERROR:
            raise Exception(f'Error in {self.integration_name}: {result.error_message}')
        if result.type == RESPONSE_TYPE.OK:
            return [], []

        df = result.data_frame
        # region clearing df from NaN values
        # recursion error appears in pandas 1.5.3 https://github.com/pandas-dev/pandas/pull/45749
        if isinstance(df, pd.Series):
            df = df.to_frame()

        try:
            df = df.replace(np.NaN, pd.NA)
        except Exception as e:
            logger.error(f"Issue with clearing DF from NaN values: {e}")

        try:
            df = df.where(pd.notnull(df), None)
        except Exception as e:
            logger.error(f"Issue with clearing DF from NaN values: {e}")
        # endregion

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in df.dtypes.items()
        ]
        data = df.to_dict(orient='records')
        return data, columns_info



import os
import urllib
from collections import OrderedDict
from typing import Any

import pandas as pd
from mindsdb_sql import parse_sql
from mindsdb_sql.parser import ast
from newsapi import NewsApiClient

from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import HandlerResponse, HandlerStatusResponse
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.config import Config


class NewsAPIArticleTable(APITable):
    def __init__(self, handler):
        super().__init__(handler)

    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where)

        params = {}

        for op, arg1, arg2 in conditions:
            if arg1 == "query":
                params["q"] = urllib.parse.quote_plus(arg2)
            elif arg1 == "sources":
                if len(arg2.split(",")) > 20:
                    raise ValueError(
                        "The number of items it sources should be 20 or less"
                    )
                else:
                    params[arg1] = arg2
            elif arg1 == "publishedAt":
                if op == "Gt" or op == "GtE":
                    params["from"] = arg2
                if op == "Lt" or op == "LtE":
                    params["to"] = arg2
                elif op == "Eq":
                    params["from"] = arg2
                    params["to"] = arg2
            else:
                params[arg1] = arg2

        if query.limit:
            if query.limit.value > 100:
                params["page"], params["page_size"] = divmod(query.limit.value, 100)
                if params["page_size"] == 0:
                    params["page_size"] = 100
            else:
                params["page_size"] = query.limit.value
                params["page"] = 1
        else:
            params["page_size"] = 100
            params["page"] = 1

        if query.order_by:
            if len(query.order_by) == 1:
                order_column = str(query.order_by[0]).split('.')[-1]
                if order_column not in ["relevancy", "publishedAt"]:
                    raise NotImplementedError("Not supported ordering by this field")
                params["sort_by"] = order_column
            else:
                raise ValueError(
                    "Multiple order by condition is not supported by the API"
                )

        selected_columns = []

        result = self.handler.call_application_api(params=params)

        if not result.empty:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    selected_columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    selected_columns.append(target.parts[-1])
                else:
                    raise ValueError(f"Unknown query target {type(target)}")

        return result[selected_columns]

    def get_columns(self) -> list:
        return [
            "author",
            "title",
            "description",
            "url",
            "urlToImage",
            "publishedAt",
            "content",
            "source_id",
            "source_name",
            "query",
            "searchIn",
            "domains",
            "excludedDomains",
        ]


class NewsAPIHandler(APIHandler):
    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.api = None
        self._tables = {}

        args = kwargs.get("connection_data", {})
        self.connection_args = {}
        handler_config = Config().get("newsAPI_handler", {})

        for k in ["api_key"]:
            if k in args:
                self.connection_args[k] = args[k]
            elif f"NEWSAPI_{k.upper()}" in os.environ:
                self.connection_args[k] = os.environ[f"NEWSAPI_{k.upper()}"]
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.is_connected = False
        self.api = self.create_connection()

        article = NewsAPIArticleTable(self)
        self._register_table("article", article)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.is_connected = False
        return self.is_connected

    def create_connection(self):
        return NewsApiClient(**self.connection_args)

    def _register_table(self, table_name: str, table_class: Any):
        self._tables[table_name] = table_class

    def get_table(self, table_name: str):
        return self._tables.get(table_name)

    def connect(self) -> HandlerStatusResponse:
        if self.is_connected is True:
            return self.api

        self.api = self.create_connection()

        self.is_connected = True
        return HandlerStatusResponse(success=True)

    def check_connection(self) -> HandlerStatusResponse:
        response = HandlerStatusResponse(False)

        try:
            self.connect()

            self.api.get_top_headlines(page_size=1, page=1)
            response.success = True

        except Exception as e:
            response.error_message = e.message

        return response

    def native_query(self, query: Any):
        ast = parse_sql(query, dialect="mindsdb")
        table = self.get_table("article")
        data = table.select(ast)
        return HandlerResponse(RESPONSE_TYPE.TABLE, data_frame=data)

    def call_application_api(
        self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        # This will implement api base on the native query
        # By processing native query to convert it to api callable parameters
        if self.is_connected is False:
            self.connect()

        pages = params.get("page", 1)
        data = []

        for page in range(1, pages + 1):
            params["page"] = page
            try:
                result = self.api.get_everything(**params)
            except Exception as e:
                raise RuntimeError(f"API call failed: {e}") 
            articles = result["articles"]
            for article in articles:
                article["source_id"] = article["source"]["id"]
                article["source_name"] = article["source"]["name"]
                del article["source"]
                article["query"] = params.get("q")
                article["searchIn"] = params.get("searchIn")
                article["domains"] = params.get("domains")
                article["excludedDomains"] = params.get("exclude_domains")
                data.append(article)

        return pd.DataFrame(data=data)


connection_args = OrderedDict(
    api_key={"type": ARG_TYPE.STR, "description": "The API key for the newsAPI API."}
)

connection_args_example = OrderedDict(api_key="knlsndlknslk")



# check if chroma_db is installed
import importlib
import shutil
import tempfile
from unittest.mock import patch

import pandas as pd
import pytest
from mindsdb_sql import parse_sql

from tests.unit.executor_test_base import BaseExecutorTest

try:
    importlib.import_module("chromadb")
    CHROMA_DB_INSTALLED = True
except ImportError:
    CHROMA_DB_INSTALLED = False


@pytest.mark.skipif(not CHROMA_DB_INSTALLED, reason="chroma_db is not installed")
class TestChromaDBHandler(BaseExecutorTest):
    def run_sql(self, sql):
        ret = self.command_executor.execute_command(parse_sql(sql, dialect="mindsdb"))

        assert ret.error_code is None
        if ret.data is not None:
            columns = [
                col.alias if col.alias is not None else col.name for col in ret.columns
            ]
            return pd.DataFrame(ret.data, columns=columns)

    @pytest.fixture(autouse=True, scope="function")
    def setup_method(self):
        super().setup_method()
        # create a chroma database under the tmp directory
        tmp_directory = tempfile.mkdtemp()
        self.run_sql(
            f"""
            CREATE DATABASE chroma_test
            WITH ENGINE = "chromadb",
            PARAMETERS = {{
                "persist_directory" : "{tmp_directory}"
            }}
            """
        )
        yield
        # Teardown code: drop the database and remove the temporary directory
        self.run_sql("DROP DATABASE chroma_test;")
        shutil.rmtree(tmp_directory)

    @pytest.mark.xfail(reason="create table for vectordatabase is not well supported")
    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_create_table(self, postgres_handler_mock):
        # create an empty table
        sql = """
            CREATE TABLE chroma_test.test_table;
        """
        self.run_sql(sql)

        # create a table with the schema definition is not allowed

        sql = """
            CREATE TABLE chroma_test.test_table (
                id int,
                metadata text,
                embedding float[]
            );
        """
        with pytest.raises(Exception):
            self.run_sql(sql)

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_create_with_select(self, postgres_handler_mock):
        df = pd.DataFrame(
            {
                "id": ["id1", "id2"],
                "content": ["this is a test", "this is a test"],
                "metadata": [{"test": "test"}, {"test": "test"}],
                "embeddings": [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]],
            }
        )

        self.set_handler(postgres_handler_mock, "pg", tables={"test_table": df})

        sql = """
        CREATE TABLE chroma_test.test_table2 (
            SELECT * FROM pg.df
        )
        """
        # this should work
        self.run_sql(sql)

    @pytest.mark.xfail(reason="drop table for vectordatabase is not working")
    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_drop_table(self, postgres_handler_mock):
        df = pd.DataFrame(
            {
                "id": ["id1", "id2"],
                "content": ["this is a test", "this is a test"],
                "metadata": [{"test": "test"}, {"test": "test"}],
                "embeddings": [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]],
            }
        )
        self.set_handler(postgres_handler_mock, "pg", tables={"test_table": df})

        # create a table
        sql = """
            CREATE TABLE chroma_test.test_table (
                SELECT * FROM pg.df
            )
        """
        self.run_sql(sql)

        # drop a table
        sql = """
            DROP TABLE chroma_test.test_table;
        """
        self.run_sql(sql)

        # drop a non existent table will raise an error
        sql = """
            DROP TABLE chroma_test.test_table2;
        """
        with pytest.raises(Exception):
            self.run_sql(sql)

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_insert_into(self, postgres_handler_mock):
        df = pd.DataFrame(
            {
                "id": ["id1", "id2", "id3"],
                "content": ["this is a test", "this is a test", "this is a test"],
                "metadata": [{"test": "test1"}, {"test": "test2"}, {"test": "test3"}],
                "embeddings": [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0], [1.0, 2.0, 3.0]],
            }
        )
        df2 = pd.DataFrame(
            {
                "id": ["id1", "id2", "id3"],
                "content": ["this is a test", "this is a test", "this is a test"],
                "metadata": [{"test": "test1"}, {"test": "test2"}, {"test": "test3"}],
                "embeddings": [
                    [1.0, 2.0, 3.0, 4.0],
                    [1.0, 2.0],
                    [1.0, 2.0, 3.0],
                ],  # different dimensions
            }
        )
        self.set_handler(postgres_handler_mock, "pg", tables={"df": df, "df2": df2})
        num_record = df.shape[0]

        # create a table
        sql = """
            CREATE TABLE chroma_test.test_table (
                SELECT * FROM pg.df
            )
        """
        self.run_sql(sql)

        # insert into a table with values
        sql = """
            INSERT INTO chroma_test.test_table (
                id,content,metadata,embeddings
            )
            VALUES (
                'some_unique_id', 'this is a test', '{"test": "test"}', '[1.0, 2.0, 3.0]'
            )
        """
        self.run_sql(sql)
        # check if the data is inserted
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE id = 'some_unique_id'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 1

        # insert without specifying id should also work
        sql = """
            INSERT INTO chroma_test.test_table (
                content,metadata,embeddings
            )
            VALUES (
                'this is a test 0', '{"test": "test"}', '[1.0, 2.0, 3.0]'
            )
        """
        self.run_sql(sql)
        # check if the data is inserted
        sql = """
            SELECT * FROM chroma_test.test_table
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == num_record + 2

        # insert into a table with a select statement
        sql = """
            INSERT INTO chroma_test.test_table (
                content,metadata,embeddings
            )
            SELECT
                content,metadata,embeddings
            FROM
                pg.df
        """
        self.run_sql(sql)
        # check if the data is inserted
        sql = """
            SELECT * FROM chroma_test.test_table
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == num_record + 3  # only one unique record was added

        # insert into a table with a select statement, but wrong columns
        with pytest.raises(Exception):
            sql = """
                INSERT INTO chroma_test.test_table
                SELECT
                    content,metadata,embeddings as wrong_column
                FROM
                    pg.df
            """
            self.run_sql(sql)

        # insert into a table with a select statement, missing metadata column
        sql = """
            INSERT INTO chroma_test.test_table
            SELECT
                content,embeddings
            FROM
                pg.df
        """
        self.run_sql(sql)

        # insert into a table with a select statement, missing embedding column, shall raise an error
        with pytest.raises(Exception):
            sql = """
                INSERT INTO chroma_test.test_table
                SELECT
                    content,metadata
                FROM
                    pg.df
            """
            self.run_sql(sql)

        # insert into a table with a select statement, with different embedding dimensions, shall raise an error
        sql = """
            INSERT INTO chroma_test.test_table
            SELECT
                content,metadata,embeddings
            FROM
                pg.df2
        """
        with pytest.raises(Exception):
            self.run_sql(sql)

        # TODO: this behavior is not consistent with chromadb doc
        # tracked in https://github.com/chroma-core/chroma/issues/1062
        # insert into a table with existing id, shall raise an error
        # sql = """
        #     INSERT INTO chroma_test.test_table (
        #         id,content,metadata,embeddings
        #     )
        #     VALUES (
        #         'id1', 'this is a test', '{"test": "test"}', '[1.0, 2.0, 3.0]'
        #     )
        # """
        # with pytest.raises(Exception):
        #     self.run_sql(sql)

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_select_from(self, postgres_handler_mock):
        df = pd.DataFrame(
            {
                "id": ["id1", "id2"],
                "content": ["this is a test", "this is a test"],
                "metadata": [{"test": "test"}, {"test": "test"}],
                "embeddings": [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]],
            }
        )
        self.set_handler(postgres_handler_mock, "pg", tables={"test_table": df})
        # create a table
        sql = """
            CREATE TABLE chroma_test.test_table (
                SELECT * FROM pg.df
            )
        """
        self.run_sql(sql)

        # query a table without any filters
        sql = """
            SELECT * FROM chroma_test.test_table
        """
        self.run_sql(sql)

        # query a table with id
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE id = 'id1'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 1

        # query a table with a search vector, without limit
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE search_vector = '[1.0, 2.0, 3.0]'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 2

        # query a table with a search vector, with limit
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE search_vector = '[1.0, 2.0, 3.0]'
            LIMIT 1
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 1

        # query a table with a metadata filter
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE `metadata.test` = 'test'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 2

        # query a table with a metadata filter and a search vector
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE `metadata.test` = 'test'
            AND search_vector = '[1.0, 2.0, 3.0]'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 2

    # @pytest.mark.xfail(reason="upsert for vectordatabase is not implemented")
    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_update(self, postgres_handler_mock):

        df = pd.DataFrame(
            {
                "id": ["id1", "id2"],
                "content": ["this is a test", "this is a test"],
                "metadata": [{"ext_id": "1"}, {"ext_id": "2"}],
                "embeddings": [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]],
            }
        )
        self.set_handler(postgres_handler_mock, "pg", tables={"test_table": df})

        # create a table
        sql = """
                    CREATE TABLE chroma_test.test_table (
                        SELECT * FROM pg.df
                    )
                """
        self.run_sql(sql)

        # updating the collection with only embeddings and not content is not allowed
        sql = """
            UPDATE chroma_test.test_table
            SET embeddings = '[3.0, 2.0, 1.0]',
             id = 'id1'
        """

        with pytest.raises(Exception):
            self.run_sql(sql)

        # updating the collection with only content and not embeddings is not allowed
        sql = """
            UPDATE chroma_test.test_table
            SET content = 'blah blah',
             id = 'id1'
        """

        with pytest.raises(Exception):
            self.run_sql(sql)

        # update multiple columns
        sql = """
            UPDATE chroma_test.test_table
            SET id = 'id1',
                embeddings = '[1.0, 2.0, 3.0]',
                content = 'this is a test'
        """
        self.run_sql(sql)
        # check if the data is updated
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE id = 'id1'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 1
        assert ret.embeddings[0] == [1.0, 2.0, 3.0]
        assert ret.content[0] == "this is a test"

        # update a table with a where clause is not allowed
        sql = """
            UPDATE chroma_test.test_table
            SET `metadata.test` = 'test2'
            WHERE search_vector = [1.0, 2.0, 3.0]
        """
        with pytest.raises(Exception):
            self.run_sql(sql)

        # update a table with all columns
        sql = """
                UPDATE chroma_test.test_table
                SET id = 'id1',
                    embeddings = '[1.0, 2.0, 3.0]',
                    content = 'this is a test',
                    `metadata.ext_id` = '1'
                """
        self.run_sql(sql)
        # check if the data is updated
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE id = 'id1'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 1
        assert ret.embeddings[0] == [1.0, 2.0, 3.0]
        assert ret.content[0] == "this is a test"
        assert ret.metadata[0] == {"ext_id": "1"}

        # update a table without providing a id is not allowed
        sql = """
            UPDATE chroma_test.test_table
            SET metadata.test = 'test3'

        """
        with pytest.raises(Exception):
            self.run_sql(sql)

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_delete(self, postgres_handler_mock):
        df = pd.DataFrame(
            {
                "id": ["id1", "id2"],
                "content": ["this is a test", "this is a test"],
                "metadata": [{"test": "test1"}, {"test": "test2"}],
                "embeddings": [[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]],
            }
        )
        self.set_handler(postgres_handler_mock, "pg", tables={"test_table": df})

        # create a table
        sql = """
            CREATE TABLE chroma_test.test_table (
                SELECT * FROM pg.df
            )
        """
        self.run_sql(sql)

        # delete from a table with a metadata filter
        sql = """
            DELETE FROM chroma_test.test_table
            WHERE `metadata.test` = 'test1'
        """
        self.run_sql(sql)
        # check if the data is deleted
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE `metadata.test` = 'test2'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 1

        # delete by id
        sql = """
            DELETE FROM chroma_test.test_table
            WHERE id = 'id2'
        """
        self.run_sql(sql)
        # check if the data is deleted
        sql = """
            SELECT * FROM chroma_test.test_table
            WHERE id = 'id2'
        """
        ret = self.run_sql(sql)
        assert ret.shape[0] == 0

        # delete from a table with a search vector filter is not allowed
        sql = """
            DELETE FROM chroma_test.test_table
            WHERE search_vector = [1.0, 2.0, 3.0]
        """
        with pytest.raises(Exception):
            self.run_sql(sql)

        # delete from a table without any filters is not allowed
        sql = """
            DELETE FROM chroma_test.test_table
        """
        with pytest.raises(Exception):
            self.run_sql(sql)



from typing import Optional
from collections import OrderedDict

import pandas as pd
from databricks import sql

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from sqlalchemy_databricks import DatabricksDialect
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

logger = log.getLogger(__name__)


class DatabricksHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Databricks statements.
    """

    name = "databricks"

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = "databricks"

        optional_parameters = [
            "session_configuration",
            "http_headers",
            "catalog",
            "schema",
        ]
        for parameter in optional_parameters:
            if parameter not in connection_data:
                connection_data[parameter] = None

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = sql.connect(
            server_hostname=self.connection_data["server_hostname"],
            http_path=self.connection_data["http_path"],
            access_token=self.connection_data["access_token"],
            session_configuration=self.connection_data["session_configuration"],
            http_headers=self.connection_data["http_headers"],
            catalog=self.connection_data["catalog"],
            schema=self.connection_data["schema"],
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(
                f'Error connecting to Databricks {self.connection_data["schema"]}, {e}!'
            )
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
                result = cursor.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result, columns=[x[0] for x in cursor.description]
                        ),
                    )

                else:
                    response = Response(RESPONSE_TYPE.OK)
                    connection.commit()
            except Exception as e:
                logger.error(
                    f'Error running query: {query} on {self.connection_data["schema"]}!'
                )
                response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender(DatabricksDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = """
            SHOW TABLES;
        """
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={"tableName": "table_name"})
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"DESCRIBE {table_name};"
        result = self.native_query(query)
        df = result.data_frame

        drop_row = df[df["col_name"] == ""].index.tolist()[0]
        df = df.iloc[: drop_row + 1]

        result.data_frame = df.rename(columns={"col_name": "column_name"})
        return result


connection_args = OrderedDict(
    server_hostname={
        "type": ARG_TYPE.STR,
        "description": "The server hostname for the cluster or SQL warehouse.",
        "required": True,
        "label": "Server_hostname",
    },
    http_path={
        "type": ARG_TYPE.STR,
        "description": "The HTTP path of the cluster or SQL warehouse.",
        "required": True,
        "label": "Http path",
    },
    access_token={
        "type": ARG_TYPE.STR,
        "description": "A Databricks personal access token for the workspace for the cluster or SQL warehouse.",
        "required": True,
        "label": "Access token",
    },
    session_configuration={
        "type": ARG_TYPE.STR,
        "description": "A dictionary of Spark session configuration parameters. This parameter is optional.",
        "required": False,
        "label": "Session configuration",
    },
    http_headers={
        "type": ARG_TYPE.STR,
        "description": "Additional (key, value) pairs to set in HTTP headers on every RPC request the client makes."
        " This parameter is optional.",
        "required": False,
        "label": "Http headers",
    },
    catalog={
        "type": ARG_TYPE.STR,
        "description": "Catalog to use for the connection. This parameter is optional.",
        "required": False,
        "label": "Catalog",
    },
    schema={
        "type": ARG_TYPE.STR,
        "description": "Schema (database) to use for the connection. This parameter is optional.",
        "required": False,
        "label": "Schema",
    },
)

connection_args_example = OrderedDict(
    server_hostname="adb-1234567890123456.7.azuredatabricks.net",
    http_path="sql/protocolv1/o/1234567890123456/1234-567890-test123",
    access_token="dapi1234567890ab1cde2f3ab456c7d89efa",
    schema="sales",
)



import time
import pandas as pd
from collections import defaultdict

from notion_client import Client

from mindsdb.utilities import log
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse,
    RESPONSE_TYPE,
)
from .notion_table import (
    NotionBlocksTable,
    NotionCommentsTable,
    NotionDatabaseTable,
    NotionPagesTable,
)

logger = log.getLogger(__name__)


class NotionHandler(APIHandler):
    name = "notion"

    def __init__(self, name: str, **kwargs):
        """constructor
        Args:
            name (str): the handler name
        """
        super().__init__(name)

        self.connection_args = kwargs.get("connection_data", {})
        self.key = "api_token"

        # set the api token from the args in create query
        if self.key in self.connection_args:
            self.connection_args[self.key] = self.connection_args[self.key]

        self.api = None
        self.is_connected = False

        notion_database_data = NotionDatabaseTable(self)
        notion_pages_data = NotionPagesTable(self)
        notion_comments_data = NotionCommentsTable(self)
        notion_blocks_data = NotionBlocksTable(self)

        self._register_table("database", notion_database_data)
        self._register_table("pages", notion_pages_data)
        self._register_table("blocks", notion_blocks_data)
        self._register_table("comments", notion_comments_data)

    def connect(self, args=None, **kwargs):
        api_token = self.connection_args[self.key]
        notion = Client(auth=api_token)
        self.is_connected = True
        return notion

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            response.error_message = (
                f"Error connecting to Notion api: {e}. Check api_token"
            )
            logger.error(response.error_message)
            response.success = False

        if response.success is False:
            self.is_connected = False

        return response

    def native_query(self, query: str = None) -> HandlerResponse:
        method_name, param = query.split("(")
        params = dict()
        # parse the query as a python function
        for map in param.strip(")").split(","):
            if map:
                k, v = map.split("=")
                params[k] = v

        df = self.call_notion_api(method_name, params)
        return HandlerResponse(RESPONSE_TYPE.TABLE, data_frame=df)

    def _apply_filters(self, data, filters):
        if not filters:
            return data

        data2 = []
        for row in data:
            add = False
            for op, key, value in filters:
                value2 = row.get(key)
                if isinstance(value, int):
                    value = str(value)

                if op in ("!=", "<>"):
                    if value == value2:
                        break
                elif op in ("==", "="):
                    if value != value2:
                        break
                elif op == "in":
                    if not isinstance(value, list):
                        value = [value]
                    if value2 not in value:
                        break
                elif op == "not in":
                    if not isinstance(value, list):
                        value = [value]
                    if value2 in value:
                        break
                else:
                    raise NotImplementedError(f"Unknown filter: {op}")
                add = True
            if add:
                data2.append(row)
        return data2

    def call_notion_api(
        self, method_name: str = None, params: dict = None, filters: list = None
    ):
        # method > table > columns
        expansions_map = {
            "database": {
                "page": ["id", "url", "properties"],
            },
            "page": {
                "properties": ["Name"],
            },
        }

        self.api = self.connect()
        # use the service as the resource to query(database, page, block, comment)
        # and query as the type of method(retrieve, list, query)
        parts = method_name.split(".")
        if len(parts) == 2:
            service, query = parts
            if service in ["databases", "pages", "blocks", "comments"]:
                method = getattr(self.api, service)
                method = getattr(method, query)
        else:
            service, children, query = parts
            if service in ["blocks"]:
                method = getattr(self.api, service)
                method = getattr(method, children)
                method = getattr(method, query)

        count_results = None
        data = []
        includes = defaultdict(list)

        max_page_size = 100
        left = None

        limit_exec_time = time.time() + 60

        if filters:
            # if we have filters: do big page requests
            params["max_results"] = max_page_size

        chunk = []
        while True:
            if time.time() > limit_exec_time:
                raise RuntimeError("Handler request timeout error")

            if count_results is not None:
                left = count_results - len(data)
                if left == 0:
                    break
                elif left < 0:
                    # got more results that we need
                    data = data[:left]
                    break

            logger.debug(f">>>notion in: {method_name}({params})")

            resp = method(**params)

            if hasattr(resp, "includes"):
                for table, records in resp.includes.items():
                    includes[table].extend([r.data for r in records])
            if resp.get("results"):
                # database and comment api has list of results
                if isinstance(resp["results"], list):
                    chunk = [r for r in resp["results"]]
            else:
                if resp.get("object") in ["page", "block"]:
                    chunk = [resp]

            if filters:
                chunk = self._apply_filters(chunk, filters)

            # limit output
            if left is not None:
                chunk = chunk[:left]

            data.extend(chunk)
            if (
                count_results is not None
                and hasattr(resp, "meta")
                and "next_token" in resp.meta
            ):
                params["next_token"] = resp.meta["next_token"]
            else:
                break

        df = pd.DataFrame(data)

        # enrich
        expansions = expansions_map.get(method_name)
        if expansions is not None:
            for table, records in includes.items():
                df_ref = pd.DataFrame(records)

                if table not in expansions:
                    continue

                for col_id in expansions[table]:
                    col = col_id[:-3]  # cut _id
                    if col_id not in df.columns:
                        continue

                    col_map = {
                        col_ref: f"{col}_{col_ref}" for col_ref in df_ref.columns
                    }
                    df_ref2 = df_ref.rename(columns=col_map)
                    df_ref2 = df_ref2.drop_duplicates(col_id)

                    df = df.merge(df_ref2, on=col_id, how="left")

        return df



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description

try:
    from .openstreetmap_handler import OpenStreetMapHandler as Handler
    
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "OpenStreetMap"
name = "openstreetmap"
type = HANDLER_TYPE.DATA
icon_path = "icon.svg"

__all__ = [
    "Handler", 
    "version", 
    "name", 
    "type", 
    "title", 
    "description",
    "import_error", 
    "icon_path",
]



from mindsdb.interfaces.storage import db


class BaseTask:

    def __init__(self, task_id, object_id):
        self.task_id = task_id
        self.object_id = object_id

    def run(self, stop_event):
        raise NotImplementedError

    def set_error(self, message):
        task_record = db.Tasks.query.get(self.task_id)
        task_record.last_error = str(message)
        db.session.commit()



import os
from typing import Optional, Dict

import cohere
import pandas as pd

from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.utilities import log


logger = log.getLogger(__name__)

class CohereHandler(BaseMLEngine):
    """
    Integration with the Cohere Python Library
    """
    name = 'cohere'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("Cohere engine requires a USING clause! Refer to its documentation for more details.")

        self.generative = True
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:

        args = self.model_storage.json_get('args')

        input_keys = list(args.keys())

        logger.info(f"Input keys: {input_keys}!")

        input_column = args['using']['column']

        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')
        
        result_df = pd.DataFrame() 

        if args['using']['task'] == 'text-summarization':
            result_df['predictions'] = df[input_column].apply(self.predict_text_summary)     

        elif args['using']['task'] == 'text-generation':
            result_df['predictions'] = df[input_column].apply(self.predict_text_generation)     

        elif args['using']['task'] == 'language-detection':
            result_df['predictions'] = df[input_column].apply(self.predict_language)     
      
        else:
            raise Exception(f"Task {args['using']['task']} is not supported!")

        result_df = result_df.rename(columns={'predictions': args['target']})
        
        return result_df


    def _get_cohere_api_key(self, args, strict=True):
        """ 
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. COHERE_API_KEY env variable
            4. cohere.api_key setting in config.json
        """ 
        # 1
        if 'api_key' in args:
            return args['api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'api_key' in connection_args:
            return connection_args['api_key']
        # 3
        api_key = os.getenv('COHERE_API_KEY')
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        openai_cfg = config.get('cohere', {})
        if 'api_key' in openai_cfg:
            return openai_cfg['api_key']

        if strict:
            raise Exception(f'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')  

    def predict_text_summary(self,text):
        """ 
        connects with cohere api to predict the summary of the input text

        """ 

        args = self.model_storage.json_get('args')

        api_key = self._get_cohere_api_key(args)
        co = cohere.Client(api_key)

        response = co.summarize(text)
        text_summary = response.summary

        return text_summary

    def predict_text_generation(self,text):
        """    
        connects with cohere api to predict the next prompt of the input text

        """
        args = self.model_storage.json_get('args')

        api_key = self._get_cohere_api_key(args)
        co = cohere.Client(api_key)

        response = co.generate(text)
        text_generated = response.generations[0].text

        return text_generated

    def predict_language(self,text):
        """    
        connects with cohere api to predict the input language 

        """
        args = self.model_storage.json_get('args')

        api_key = self._get_cohere_api_key(args)
        co = cohere.Client(api_key)

        response = co.detect_language([text])
        language_detected = response.results[0].language_name

        return language_detected
    



from collections import OrderedDict
from typing import List, Optional

import pandas as pd
import json
import xata
from xata.helpers import BulkProcessor

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class XataHandler(VectorStoreHandler):
    """This handler handles connection and execution of the Xata statements."""

    name = "xata"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self._connection_data = kwargs.get("connection_data")
        self._client_config = {
            "db_url": self._connection_data.get("db_url"),
            "api_key": self._connection_data.get("api_key"),
        }
        self._create_table_params = {
            "dimension": self._connection_data.get("dimension", 8),
        }
        self._select_params = {
            "similarity_function": self._connection_data.get("similarity_function", "cosineSimilarity"),
        }
        self._client = None
        self.is_connected = False
        self.connect()

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """Connect to a Xata database."""
        if self.is_connected is True:
            return self._client
        try:
            self._client = xata.XataClient(**self._client_config)
            self.is_connected = True
            return self._client
        except Exception as e:
            logger.error(f"Error connecting to Xata client: {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the database connection."""
        if self.is_connected is False:
            return
        self._client = None
        self.is_connected = False

    def check_connection(self):
        """Check the connection to the Xata database."""
        response_code = StatusResponse(False)
        need_to_close = self.is_connected is False
        # NOTE: no direct way to test this
        # try getting the user, if it fails, it means that we are not connected
        try:
            resp = self._client.users().get()
            if not resp.is_success():
                raise Exception(resp["message"])
            response_code.success = True
        except Exception as e:
            logger.error(f"Error connecting to Xata: {e}!")
            response_code.error_message = str(e)
        finally:
            if response_code.success is True and need_to_close:
                self.disconnect()
            if response_code.success is False and self.is_connected is True:
                self.is_connected = False
        return response_code

    def create_table(self, table_name: str, if_not_exists=True) -> HandlerResponse:
        """Create a table with the given name in the Xata database."""
        try:
            resp = self._client.table().create(table_name)
            if not resp.is_success():
                raise Exception(f"Unable to create table {table_name}: {resp['message']}")
            resp = self._client.table().set_schema(
                table_name=table_name,
                payload={
                    "columns": [
                        {
                            "name": "embeddings",
                            "type": "vector",
                            "vector": {"dimension": self._create_table_params["dimension"]}
                        },
                        {"name": "content", "type": "text"},
                        {"name": "metadata", "type": "json"},
                    ]
                }
            )
            if not resp.is_success():
                raise Exception(f"Unable to change schema of table {table_name}: {resp['message']}")
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"{e}",
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def drop_table(self, table_name: str, if_exists=True) -> HandlerResponse:
        """Delete a table from the Xata database."""
        try:
            resp = self._client.table().delete(table_name)
            if not resp.is_success():
                raise Exception(f"Unable to delete table: {resp['message']}")
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"{e}",
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def get_columns(self, table_name: str) -> HandlerResponse:
        """Get columns of the given table"""
        # Vector stores have predefined columns
        try:
            # But at least try to see if the table is valid
            resp = self._client.table().get_columns(table_name)
            if not resp.is_success():
                raise Exception(f"Error getting columns: {resp['message']}")
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"{e}",
            )
        return super().get_columns(table_name)

    def get_tables(self) -> HandlerResponse:
        """Get the list of tables in the Xata database."""
        try:
            table_names = pd.DataFrame(
                columns=["TABLE_NAME"],
                data=[table_data["name"] for table_data in self._client.branch().get_details()["schema"]["tables"]],
            )
            return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=table_names)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error getting list of tables: {e}",
            )

    def insert(self, table_name: str, data: pd.DataFrame, columns: List[str] = None) -> HandlerResponse:
        """ Insert data into the Xata database. """
        if columns:
            data = data[columns]
        # Convert to records
        data = data.to_dict("records")
        # Convert metadata to json
        for row in data:
            if "metadata" in row:
                row["metadata"] = json.dumps(row["metadata"])
        if len(data) > 1:
            # Bulk processing
            try:
                bp = BulkProcessor(self._client, throw_exception=True)
                bp.put_records(table_name, data)
                bp.flush_queue()
            except Exception as e:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Error inserting data into '{table_name}': {e}",
                )
        elif len(data) == 0:
            # Skip
            return Response(resp_type=RESPONSE_TYPE.OK)
        elif "id" in data[0] and TableField.ID.value in columns:
            # If id present
            id = data[0]["id"]
            rest_of_data = data[0].copy()
            del rest_of_data["id"]
            try:
                resp = self._client.records().insert_with_id(
                    table_name=table_name,
                    record_id=id,
                    payload=rest_of_data,
                    create_only=True,
                    columns=columns
                )
                if not resp.is_success():
                    raise Exception(resp["message"])
            except Exception as e:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Error inserting data into '{table_name}': {e}",
                )
        else:
            # If id not present
            try:
                resp = self._client.records().insert(
                    table_name=table_name,
                    payload=data[0],
                    columns=columns
                )
                if not resp.is_success():
                    raise Exception(resp["message"])
            except Exception as e:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Error inserting data into '{table_name}': {e}",
                )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def update(self, table_name: str, data: pd.DataFrame, columns: List[str] = None) -> HandlerResponse:
        """Update data in the Xata database."""
        # Not supported
        return super().update(table_name, data, columns)

    def _get_xata_operator(self, operator: FilterOperator) -> str:
        """Translate SQL operator to oprator understood by Xata filter language."""
        mapping = {
            FilterOperator.EQUAL: "$is",
            FilterOperator.NOT_EQUAL: "$isNot",
            FilterOperator.LESS_THAN: "$lt",
            FilterOperator.LESS_THAN_OR_EQUAL: "$le",
            FilterOperator.GREATER_THAN: "$gt",
            FilterOperator.GREATER_THAN_OR_EQUAL: "$gte",
            FilterOperator.LIKE: "$pattern",
        }
        if operator not in mapping:
            raise Exception(f"Operator '{operator}' is not supported!")
        return mapping[operator]

    def _translate_non_vector_conditions(self, conditions: List[FilterCondition]) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects a dict that can be used by Xata for filtering.
        E.g.,
        [
            FilterCondition(
                column="metadata.price",
                op=FilterOperator.LESS_THAN,
                value=100,
            ),
            FilterCondition(
                column="metadata.price",
                op=FilterOperator.GREATER_THAN,
                value=10,
            )
        ]
        -->
        {
            "metadata->price" {
                "$gt": 10,
                "$lt": 100
            },
        }
        """
        if not conditions:
            return None
        # Translate metadata columns
        for condition in conditions:
            if condition.column.startswith(TableField.METADATA.value):
                condition.column = condition.column.replace(".", "->")
        # Generate filters
        filters = {}
        for condition in conditions:
            # Skip search vector condition
            if condition.column == TableField.SEARCH_VECTOR.value:
                continue
            current_filter = original_filter = {}
            # Special case LIKE: needs pattern translation
            if condition.op == FilterOperator.LIKE:
                condition.value = condition.value.replace("%", "*").replace("_", "?")
            # Generate substatment
            current_filter[condition.column] = {self._get_xata_operator(condition.op): condition.value}
            # Check for conflicting and insert
            for key in original_filter:
                if key in filters:
                    filters[key] = {**filters[key], **original_filter[key]}
                else:
                    filters = {**filters, **original_filter}
        return filters if filters else None

    def select(self, table_name: str, columns: List[str] = None, conditions: List[FilterCondition] = None, offset: int = None, limit: int = None) -> HandlerResponse:
        """Run general query or a vector similarity search and return results."""
        if not columns:
            columns = [col["name"] for col in self.SCHEMA]
        # Generate filter conditions
        filters = self._translate_non_vector_conditions(conditions)
        # Check for search vector
        search_vector = (
            []
            if conditions is None
            else [
                condition.value
                for condition in conditions
                if condition.column == TableField.SEARCH_VECTOR.value
            ]
        )
        if len(search_vector) > 0:
            search_vector = search_vector[0]
        else:
            search_vector = None
        # Search
        results_df = pd.DataFrame(columns)
        if search_vector is not None:
            # Similarity
            try:
                params = {
                    "queryVector": search_vector,
                    "column": TableField.EMBEDDINGS.value,
                    "similarityFunction": self._select_params["similarity_function"]
                }
                if filters:
                    params["filter"] = filters
                if limit:
                    params["size"] = limit
                results = self._client.data().vector_search(table_name, params)
                # Check for errors
                if not results.is_success():
                    raise Exception(results["message"])
                # Convert result
                results_df = pd.DataFrame.from_records(results["records"])
                if "xata" in results_df.columns:
                    results_df["xata"] = results_df["xata"].apply(lambda x: x["score"])
                    results_df.rename({"xata": TableField.DISTANCE.value}, axis=1, inplace=True)
            except Exception as e:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Unable to run similarity search: {e}",
                )
        else:
            # General get query
            try:
                params = {
                    "columns": columns if columns else [],
                }
                if filters:
                    params["filter"] = filters
                if limit or offset:
                    params["page"] = {}
                    if limit:
                        params["page"]["size"] = limit
                    if offset:
                        params["page"]["offset"] = offset
                results = self._client.data().query(table_name, params)
                # Check for errors
                if not results.is_success():
                    raise Exception(results["message"])
                # Convert result
                results_df = pd.DataFrame.from_records(results["records"])
                if "xata" in results_df.columns:
                    results_df.drop(["xata"], axis=1, inplace=True)
            except Exception as e:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Unable to run general SELECT query : {e}",
                )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=results_df)

    def delete(self, table_name: str, conditions: List[FilterCondition] = None) -> HandlerResponse:
        ids = []
        for condition in conditions:
            if condition.op == FilterOperator.EQUAL:
                ids.append(condition.value)
            else:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message="You can only delete using '=' operator ID one at a time!",
                )
        try:
            for id in ids:
                resp = self._client.records().delete(table_name, id)
                if not resp.is_success():
                    raise Exception(resp["message"])
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to run DELETE records: {e}",
            )
        return Response(resp_type=RESPONSE_TYPE.OK)


connection_args = OrderedDict(
    db_url={
        "type": ARG_TYPE.STR,
        "description": "Xata database url with region, database and, optionally the branch information",
        "required": True,
    },
    api_key={
        "type": ARG_TYPE.STR,
        "description": "personal Xata API key",
        "required": True,
    },
    dimension={
        "type": ARG_TYPE.INT,
        "description": "default dimension of embeddings vector used to create table when using create (default=8)",
        "required": False,
    },
    similarity_function={
        "type": ARG_TYPE.STR,
        "description": "similarity function to use for vector searches (default=cosineSimilarity)",
        "required": False,
    }
)

connection_args_example = OrderedDict(
    db_url="https://...",
    api_key="abc_def...",
    dimension=8,
    similarity_function="l1"
)



import os

import pandas as pd

from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from mindsdb.integrations.libs.api_handler import APIHandler

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from .tripadvisor_table import SearchLocationTable
from .tripadvisor_table import LocationDetailsTable
from .tripadvisor_table import ReviewsTable
from .tripadvisor_table import PhotosTable
from .tripadvisor_table import NearbyLocationTable
from .tripadvisor_api import TripAdvisorAPI
from .tripadvisor_api import TripAdvisorAPICall

logger = log.getLogger(__name__)


class TripAdvisorHandler(APIHandler):
    """A class for handling connections and interactions with the TripAdvisor Content API.

    Attributes:
        api_key (str): The unique API key to access Tripadvisor content.
        api (TripAdvisorAPI): The `TripAdvisorAPI` object for checking the connection to the TripAdvisor API.
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        args = kwargs.get("connection_data", {})
        self._tables = {}

        self.connection_args = {}
        handler_config = Config().get("tripadvisor_handler", {})
        for k in ["api_key"]:
            if k in args:
                self.connection_args[k] = args[k]
            elif f"TRIPADVISOR_{k.upper()}" in os.environ:
                self.connection_args[k] = os.environ[f"TRIPADVISOR_{k.upper()}"]
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]

        self.api = None
        self.is_connected = False

        tripAdvisor = SearchLocationTable(self)
        self._register_table("searchLocationTable", tripAdvisor)

        tripAdvisorLocationDetails = LocationDetailsTable(self)
        self._register_table("locationDetailsTable", tripAdvisorLocationDetails)

        tripAdvisorReviews = ReviewsTable(self)
        self._register_table("reviewsTable", tripAdvisorReviews)

        tripAdvisorPhotos = PhotosTable(self)
        self._register_table("photosTable", tripAdvisorPhotos)

        tripAdvisorNearbyLocation = NearbyLocationTable(self)
        self._register_table("nearbyLocationTable", tripAdvisorNearbyLocation)

    def connect(self, api_version=2):
        """Check the connection with TripAdvisor API"""

        if self.is_connected is True:
            return self.api

        self.api = TripAdvisorAPI(api_key=self.connection_args["api_key"])

        self.is_connected = True
        return self.api

    def check_connection(self) -> StatusResponse:
        """This function evaluates if the connection is alive and healthy"""
        response = StatusResponse(False)

        try:
            api = self.connect()

            # make a random http call with searching a location.
            #   it raises an error in case if auth is not success and returns not-found otherwise
            api.connectTripAdvisor()
            response.success = True

        except Exception as e:
            response.error_message = f"Error connecting to TripAdvisor api: {e}"
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def call_tripadvisor_searchlocation_api(
        self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        """It processes the JSON data from the call and transforms it into pandas.Dataframe"""
        if self.is_connected is False:
            self.connect()

        locations = self.api.getTripAdvisorData(
            TripAdvisorAPICall.SEARCH_LOCATION, **params
        )
        result = []

        for loc in locations:
            data = {
                "location_id": loc.get("location_id"),
                "name": loc.get("name"),
                "distance": loc.get("distance"),
                "rating": loc.get("rating"),
                "bearing": loc.get("bearing"),
                "street1": loc.get("address_obj").get("street1"),
                "street2": loc.get("address_obj").get("street2"),
                "city": loc.get("address_obj").get("city"),
                "state": loc.get("address_obj").get("state"),
                "country": loc.get("address_obj").get("country"),
                "postalcode": loc.get("address_obj").get("postalcode"),
                "address_string": loc.get("address_obj").get("address_string"),
                "phone": loc.get("address_obj").get("phone"),
                "latitude": loc.get("address_obj").get("latitude"),
                "longitude": loc.get("address_obj").get("longitude"),
            }
            result.append(data)
        result = pd.DataFrame(result)
        return result

    def call_tripadvisor_location_details_api(
        self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        """It processes the JSON data from the call and transforms it into pandas.Dataframe"""
        if self.is_connected is False:
            self.connect()

        loc = self.api.getTripAdvisorData(TripAdvisorAPICall.LOCATION_DETAILS, **params)
        result = []

        data = {
            "location_id": loc.get("location_id"),
            "name": loc.get("name"),
            "distance": loc.get("distance"),
            "rating": loc.get("rating"),
            "bearing": loc.get("bearing"),
            "street1": loc.get("address_obj").get("street1"),
            "street2": loc.get("address_obj").get("street2"),
            "city": loc.get("address_obj").get("city"),
            "state": loc.get("address_obj").get("state"),
            "country": loc.get("address_obj").get("country"),
            "postalcode": loc.get("address_obj").get("postalcode"),
            "address_string": loc.get("address_obj").get("address_string"),
            "phone": loc.get("address_obj").get("phone"),
            "latitude": loc.get("address_obj").get("latitude"),
            "longitude": loc.get("address_obj").get("longitude"),
            "web_url": loc.get("web_url"),
            "timezone": loc.get("timezone"),
            "email": loc.get("email"),
            "website": loc.get("website"),
            "write_review": loc.get("write_review"),
            "ranking_data": str(loc.get("ranking_data")),
            "rating_image_url": loc.get("rating_image_url"),
            "num_reviews": loc.get("num_reviews"),
            "review_rating_count": loc.get("review_rating_count"),
            "subratings": loc.get("subratings"),
            "photo_count": loc.get("photo_count"),
            "see_all_photos": loc.get("see_all_photos"),
            "price_level": loc.get("price_level"),
            "parent_brand": loc.get("parent_brand"),
            "brand": loc.get("brand"),
            "ancestors": str(loc.get("ancestors")),
            "periods": str(loc.get("hours").get("periods"))
            if loc.get("hours") is not None
            else None,
            "weekday": str(loc.get("hours").get("weekday_text"))
            if loc.get("weekday") is not None
            else None,
            "amenities": str(loc.get("amenities")),
            "features": str(loc.get("features")),
            "cuisines": str(loc.get("cuisine")),
            "styles": str(loc.get("styles")),
            "neighborhood_info": str(loc.get("neighborhood_info")),
            "awards": str(loc.get("awards")),
            "trip_types": str(loc.get("trip_types")),
            "groups": str(loc.get("groups")),
        }

        result.append(data)

        result = pd.DataFrame(result)
        return result

    def call_tripadvisor_reviews_api(
        self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        """It processes the JSON data from the call and transforms it into pandas.Dataframe"""
        if self.is_connected is False:
            self.connect()

        locations = self.api.getTripAdvisorData(TripAdvisorAPICall.REVIEWS, **params)
        result = []

        for loc in locations:
            data = {
                "id": loc.get("id"),
                "lang": loc.get("lang"),
                "location_id": loc.get("location_id"),
                "published_date": loc.get("published_date"),
                "rating": loc.get("rating"),
                "helpful_votes": loc.get("helpful_votes"),
                "rating_image_url": loc.get("rating_image_url"),
                "url": loc.get("url"),
                "trip_type": loc.get("trip_type"),
                "travel_date": loc.get("travel_date"),
                "text_review": loc.get("text"),
                "title": loc.get("title"),
                "owner_response": loc.get("owner_response"),
                "is_machine_translated": loc.get("is_machine_translated"),
                "user": str(loc.get("user")),
                "subratings": str(loc.get("subratings")),
            }
            result.append(data)
        result = pd.DataFrame(result)
        return result

    def call_tripadvisor_photos_api(
        self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        """It processes the JSON data from the call and transforms it into pandas.Dataframe"""
        if self.is_connected is False:
            self.connect()

        locations = self.api.getTripAdvisorData(TripAdvisorAPICall.PHOTOS, **params)
        result = []

        for loc in locations:
            data = {
                "id": loc.get("id"),
                "is_blessed": loc.get("is_blessed"),
                "album": loc.get("album"),
                "caption": loc.get("caption"),
                "published_date": loc.get("published_date"),
                "images": str(loc.get("images")),
                "source": str(loc.get("source")),
                "user": str(loc.get("user")),
            }
            result.append(data)
        result = pd.DataFrame(result)
        return result

    def call_tripadvisor_nearby_location_api(
        self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        """It processes the JSON data from the call and transforms it into pandas.Dataframe"""
        if self.is_connected is False:
            self.connect()

        locations = self.api.getTripAdvisorData(TripAdvisorAPICall.NEARBY_SEARCH, **params)
        result = []

        for loc in locations:
            data = {
                "location_id": loc.get("location_id"),
                "name": loc.get("name"),
                "distance": loc.get("distance"),
                "rating": loc.get("rating"),
                "bearing": loc.get("bearing"),
                "address_obj": str(loc.get("address_obj")),
            }
            result.append(data)
        result = pd.DataFrame(result)
        return result



"""
Utility functions used in the 'Bring Your Own Model' (BYOM) engine.

These functions interact with interfaces (stdin, stdout), python files, and the actual BYOM engine.

In particular, they:
    - Wrap and run python code in separate python proceess.
    - Communicate with parent process throughout stdin/out using pickle to serialize objects.


The flow is as follows:

    1. Receive module code, method with parameters and stored attributes from parent process
    2. A python class object is created from the code
    3. Class is instanced and filled with stored attributes
    4. A calls to the chosen method of the class is performed with any relevant parameters that were passed
    5. Response is generated, appropriately packaged and sent to stdout
    6. Exit
"""

import io
import sys
import pickle
import inspect
from enum import Enum

import pandas as pd


class BYOM_METHOD(Enum):
    CHECK = 1
    TRAIN = 2
    PREDICT = 3
    FINETUNE = 4
    DESCRIBE = 5


def pd_encode(df):
    return df.to_parquet(engine='pyarrow')


def pd_decode(encoded):
    fd = io.BytesIO()
    fd.write(encoded)
    fd.seek(0)
    return pd.read_parquet(fd, engine='pyarrow')


def encode(obj):
    return pickle.dumps(obj, protocol=5)


def decode(encoded):
    return pickle.loads(encoded)


def return_output(obj):
    # read stdin
    encoded = encode(obj)
    with open(1, 'wb') as fd:
        fd.write(encoded)
    sys.exit(0)


def get_input():
    # write to stdout
    with open(0, 'rb') as fd:
        encoded = fd.read()
        obj = decode(encoded)
    return obj


def import_string(code, module_name='model'):
    # import string as python module

    import types
    module = types.ModuleType(module_name)

    exec(code, module.__dict__)
    # sys.modules['my_module'] = module
    return module


def find_model_class(module):
    # find the first class that contents predict and train methods
    for _, klass in inspect.getmembers(module, inspect.isclass):
        funcs = [
            name
            for name, _ in inspect.getmembers(klass, inspect.isfunction)
        ]
        if 'predict' in funcs and 'train' in funcs:
            return klass


def main():
    # replace print output to stderr
    sys.stdout = sys.stderr

    params = get_input()

    method = BYOM_METHOD(params['method'])
    code = params['code']

    module = import_string(code)

    model_class = find_model_class(module)

    if method == BYOM_METHOD.CHECK:
        model = model_class()

        if not hasattr(model, 'train'):
            raise RuntimeError('Model class has to have "train" method')

        if not hasattr(model, 'predict'):
            raise RuntimeError('Model class has to have "predict" method')

        return_output(True)

    if method == BYOM_METHOD.TRAIN:
        df = pd_decode(params['df'])
        to_predict = params['to_predict']
        args = params['args']
        model = model_class()

        call_args = [df, to_predict]
        if args:
            call_args.append(args)
        model.train(*call_args)

        # return model
        data = model.__dict__

        model_state = encode(data)
        return_output(model_state)

    elif method == BYOM_METHOD.PREDICT:
        model_state = params['model_state']
        df = pd_decode(params['df'])
        args = params['args']

        model = model_class()
        model.__dict__ = decode(model_state)

        call_args = [df]
        if args:
            call_args.append(args)
        res = model.predict(*call_args)
        return_output(pd_encode(res))

    elif method == BYOM_METHOD.FINETUNE:
        model_state = params['model_state']
        df = pd_decode(params['df'])
        args = params['args']

        model = model_class()
        model.__dict__ = decode(model_state)

        call_args = [df]
        if args:
            call_args.append(args)

        model.finetune(*call_args)

        # return model
        data = model.__dict__
        model_state = encode(data)
        return_output(model_state)

    elif method == BYOM_METHOD.DESCRIBE:
        model_state = params['model_state']
        model = model_class()
        model.__dict__ = decode(model_state)
        try:
            df = model.describe(params.get('attribute'))
        except Exception:
            return_output(pd_encode(pd.DataFrame()))
        return_output(pd_encode(df))

    raise NotImplementedError(method)


if __name__ == '__main__':
    main()



import pandas as pd
from typing import Dict, List

from rocketchat_API.rocketchat import RocketChat

from mindsdb.integrations.handlers.rocket_chat_handler.rocket_chat_tables import (
    ChannelMessagesTable, ChannelsTable, DirectsTable, DirectMessagesTable, UsersTable)
from mindsdb.integrations.libs.api_handler import APIChatHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
from mindsdb_sql import parse_sql

logger = log.getLogger(__name__)

class RocketChatHandler(APIChatHandler):
    """A class for handling connections and interactions with the Rocket Chat API.

    Attributes:
        username (str): Rocket Chat username to use for authentication.
        password (str): Rocket Chat username to use for authentication.
        auth_token (str): Rocket Chat authorization token to use for all API requests.
        auth_user_id (str): Rocket Chat user ID to associate with all API requests
        domain (str): Path to Rocket Chat domain to use (e.g. https://mindsdb.rocket.chat).
        client (RocketChatClient): The `RocketChatClient` object for interacting with the Rocket Chat API.
        is_connected (bool): Whether or not the API client is connected to the Rocket Chat API.
    """

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        self.username = None
        self.password = None
        self.auth_token = None
        self.auth_user_id = None
        self.domain = None

        args = kwargs.get('connection_data', {})
        if 'domain' not in args:
            raise ValueError('Must include Rocket Chat "domain" to read and write messages')
        self.domain = args['domain']

        if 'token' in args and 'user_id' in args:
            self.auth_token = args['token']
            self.auth_user_id = args['user_id']
        elif 'username' in args and 'password' in args:
            self.username = args['username']
            self.password = args['password']
        else:
            raise ValueError('Need "token" and "user_id", or "username" and "password" to connect to Rocket Chat')

        self.client = None
        self.is_connected = False

        self._register_table('channels', ChannelsTable(self))

        self._register_table('channel_messages', ChannelMessagesTable(self))

        self._register_table('directs', DirectsTable(self))

        self._register_table('direct_messages', DirectMessagesTable(self))

        self._register_table('users', UsersTable(self))

    def get_chat_config(self):
        params = {
            'polling': {
                'type': 'message_count',
                'table': 'directs',
                'chat_id_col': '_id',
                'count_col': 'msgs'
            },
            'chat_table': {
                'name': 'direct_messages',
                'chat_id_col': 'room_id',
                'username_col': 'username',
                'text_col': 'text',
                'time_col': 'sent_at',
            }
        }
        return params

    def get_my_user_name(self):
        info = self.call_api('me')
        return info['username']

    def connect(self):
        """Creates a new Rocket Chat API client if needed and sets it as the client to use for requests.

        Returns newly created Rocket Chat API client, or current client if already set.
        """
        if self.is_connected and self.client is not None:
            return self.client

        self.client = RocketChat(
            user=self.username,
            password=self.password,
            auth_token=self.auth_token,
            user_id=self.auth_user_id,
            server_url=self.domain
        )

        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        """Checks connection to Rocket Chat API by sending a ping request.

        Returns StatusResponse indicating whether or not the handler is connected.
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Rocket Chat API: {e}!')
            response.error_message = e

        if response.success is False:
            self.is_connected = False
        return response

    def native_query(self, query: str = None) -> Response:
        ast = parse_sql(query, dialect='mindsdb')
        return self.query(ast)


    def call_api(self, method_name: str = None, *args, **kwargs) -> pd.DataFrame:
        """Calls the Rocket Chat API method with the given params.

        Returns results as a pandas DataFrame.

        Args:
            method_name (str): Method name to call
            params (Dict): Params to pass to the API call
        """
        client = self.connect()

        method = getattr(client, method_name)

        messages_response = method(*args, **kwargs)

        if not messages_response.ok:
            messages_response.raise_for_status()
        return messages_response.json()



from copy import deepcopy

import pandas as pd

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import (
    BinaryOperation,
    Identifier,
    Constant,
    Update,
    Select,
    Delete,
)

from mindsdb.utilities.exception import EntityNotExistsError
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import SQLQuery
from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.interfaces.query_context.context_controller import query_context_controller


class ProjectDataNode(DataNode):
    type = 'project'

    def __init__(self, project, integration_controller, information_schema):
        self.project = project
        self.integration_controller = integration_controller
        self.information_schema = information_schema

    def get_type(self):
        return self.type

    def get_tables(self):
        tables = self.project.get_tables()
        table_types = {
            'table': 'BASE TABLE',
            'model': 'MODEL',
            'view': 'VIEW'
        }
        tables = [
            {
                'TABLE_NAME': key,
                'TABLE_TYPE': table_types.get(val['type'])
            }
            for key, val in tables.items()
        ]
        result = [TablesRow.from_dict(row) for row in tables]
        return result

    def has_table(self, table_name):
        tables = self.project.get_tables()
        return table_name in tables

    def get_table_columns(self, table_name):
        return self.project.get_columns(table_name)

    def predict(self, model_name: str, data, version=None, params=None):
        model_metadata = self.project.get_model(model_name)
        if model_metadata is None:
            raise Exception(f"Can't find model '{model_name}'")
        model_metadata = model_metadata['metadata']
        if model_metadata['update_status'] == 'available':
            raise Exception(f"model '{model_name}' is obsolete and needs to be updated. Run 'RETRAIN {model_name};'")
        handler = self.integration_controller.get_handler(model_metadata['engine_name'])
        return handler.predict(model_name, data, project_name=self.project.name, version=version, params=params)

    def query(self, query=None, native_query=None, session=None):
        if query is None and native_query is not None:
            query = parse_sql(native_query, dialect='mindsdb')

        if isinstance(query, Update):
            query_table = query.table.parts[0].lower()
            kb_table = session.kb_controller.get_table(query_table, self.project.id)
            if kb_table:
                # this is the knowledge db
                kb_table.update_query(query)
                return pd.DataFrame(), []

            raise NotImplementedError(f"Can't update object: {query_table}")

        elif isinstance(query, Delete):
            query_table = query.table.parts[0].lower()
            kb_table = session.kb_controller.get_table(query_table, self.project.id)
            if kb_table:
                # this is the knowledge db
                kb_table.delete_query(query)
                return pd.DataFrame(), []

            raise NotImplementedError(f"Can't delete object: {query_table}")

        elif isinstance(query, Select):
            # region is it query to 'models' or 'models_versions'?
            query_table = query.from_table.parts[0].lower()
            # region FIXME temporary fix to not broke queries to 'mindsdb.models'. Can be deleted it after 1.12.2022
            if query_table == 'predictors':
                query.from_table.parts[0] = 'models'
                query_table = 'models'
            # endregion
            if query_table in ('models', 'models_versions', 'jobs', 'jobs_history', 'mdb_triggers', 'chatbots', 'skills', 'agents'):
                new_query = deepcopy(query)
                project_filter = BinaryOperation('=', args=[
                    Identifier('project'),
                    Constant(self.project.name)
                ])
                if new_query.where is None:
                    new_query.where = project_filter
                else:
                    new_query.where = BinaryOperation('and', args=[
                        new_query.where,
                        project_filter
                    ])
                data, columns_info = self.information_schema.query(new_query)
                return data, columns_info
            # endregion

            # other table from project

            if self.project.get_view(query_table):
                # this is the view

                view_meta = self.project.query_view(query)

                query_context_controller.set_context('view', view_meta['id'])

                try:
                    sqlquery = SQLQuery(
                        view_meta['query_ast'],
                        session=session
                    )
                    result = sqlquery.fetch(view='dataframe')

                finally:
                    query_context_controller.release_context('view', view_meta['id'])

                if result['success'] is False:
                    raise Exception(f"Cant execute view query: {view_meta['query_ast']}")
                df = result['result']

                df = query_df(df, query)

                columns_info = [
                    {
                        'name': k,
                        'type': v
                    }
                    for k, v in df.dtypes.items()
                ]

                return df.to_dict(orient='records'), columns_info

            kb_table = session.kb_controller.get_table(query_table, self.project.id)
            if kb_table:
                # this is the knowledge db
                df = kb_table.select_query(query)
                columns_info = [
                    {
                        'name': k,
                        'type': v
                    }
                    for k, v in df.dtypes.items()
                ]

                return df.to_dict(orient='records'), columns_info

            raise EntityNotExistsError(f"Can't select from {query_table} in project")
        else:
            raise NotImplementedError(f"Query not supported {query}")

    def create_table(self, table_name: Identifier, result_set, is_replace=False, **kwargs):
        # is_create - create table
        # is_replace - drop table if exists
        # is_create==False and is_replace==False: just insert

        from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
        session = SessionController()

        table_name = table_name.parts[-1]
        kb_table = session.kb_controller.get_table(table_name, self.project.id)
        if kb_table:
            # this is the knowledge db
            if is_replace:
                kb_table.clear()

            df = result_set.to_df()
            return kb_table.insert(df)
        raise NotImplementedError(f"Cant create table {table_name}")



"""This module contains some useful features which
help to increase solution stability in network layer"""

import time


def sending_attempts(exception_type=Exception, attempts_number=3, delay=0.5):
    """Decorator factory for functions where several call attempts may require for success"""
    def decorator(func):
        def wrapper(*args, **kwargs):
            to_raise = None
            res = None
            for _ in range(attempts_number):
                try:
                    res = func(*args, **kwargs)
                    break
                # Save catched exception
                except Exception as e:
                    to_raise = e
                    time.sleep(delay)
            # Else executed if there was no loop breaks
            else:
                raise exception_type(f"{type(to_raise)}: {to_raise}") from to_raise  # noqa: pylint: disable=raising-bad-type
            return res
        return wrapper
    return decorator



from typing import List, Optional

from pydantic import BaseModel, Extra, root_validator

from mindsdb.integrations.handlers.utilities.validation_utilities import ParameterValidationUtilities


class TwelveLabsHandlerConfig(BaseModel):
    "Configuration for TwelveLabsHandler."

    index_name: str
    engine_id: Optional[str] = None
    api_key: Optional[str] = None
    index_options: List[str]
    addons: List[str] = []
    video_urls: Optional[List[str]] = None
    video_urls_col: Optional[str] = None
    video_files: Optional[List[str]] = None
    video_files_col: Optional[str] = None
    task: str = None
    search_options: Optional[List[str]] = None

    class Config:
        extra = Extra.forbid

    @root_validator(pre=True, allow_reuse=True)
    def check_param_typos(cls, values):
        """Check if there are any typos in the parameters."""

        ParameterValidationUtilities.validate_parameter_spelling(cls, values)

    @root_validator(allow_reuse=True)
    def check_for_video_urls_or_video_files(cls, values):
        """Check if video_urls or video_files have been provided."""

        video_urls = values.get("video_urls")
        video_urls_col = values.get("video_urls_col")
        video_files = values.get("video_files")
        video_files_col = values.get("video_files_col")

        if not video_urls and not video_files and not video_urls_col and not video_files_col:
            raise ValueError(
                "Neither video_urls, video_files, video_urls_col nor video_files_col have been provided. Please provide one of them."
            )

        return values

    @root_validator(allow_reuse=True)
    def check_for_task(cls, values):
        """Check if task has been provided along with the other relevant parameters for each task."""

        task = values.get("task")

        if task == "search":
            search_options = values.get("search_options")
            if not search_options:
                raise ValueError(
                    "search_options have not been provided. Please provide search_options."
                )

            # search options should be a subset of index options
            index_options = values.get("index_options")
            if not set(search_options).issubset(set(index_options)):
                raise ValueError(
                    "search_options should be a subset of index_options."
                )

        else:
            raise ValueError(
                f"task {task} is not supported. Please provide a valid task."
            )

        return values



from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from mindsdb.integrations.handlers.slack_handler import SlackHandler
import unittest
from unittest.mock import patch
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb_sql import parse_sql

class SlackHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.handler = SlackHandler('test-slack-token')

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    @patch('mindsdb.integrations.handlers.slack_handler.slack_handler.requests.post')
    def test_1_send_message(self, mock_post):
        mock_post.return_value.status_code = 200
        self.handler.send_message("#test-channel", "Test message")
        mock_post.assert_called_with(
            'https://slack.com/api/chat.postMessage',
            headers={'Authorization': 'Bearer test-slack-token'},
            json={'channel': '#test-channel', 'text': 'Test message'}
        )

if __name__ == '__main__':
    unittest.main()



import unittest
from mindsdb.integrations.handlers.scylla_handler.scylla_handler import ScyllaHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class ScyllaHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "",
                "port": "9042",
                "user": "",
                "password": "",
                "keyspace": "test_data"
            }
        }
        cls.handler = ScyllaHandler('test_scylla_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_1_native_query_show_keyspaces(self):
        dbs = self.handler.native_query("DESC KEYSPACES;")
        assert dbs['type'] is not RESPONSE_TYPE.ERROR

    def test_2_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls['type'] is not RESPONSE_TYPE.ERROR

    def test_3_describe_table(self):
        described = self.handler.get_columns("home_rentals")
        assert described['type'] is RESPONSE_TYPE.TABLE

    def test_4_select_query(self):
        query = "SELECT * FROM home_rentals WHERE 'id'='3712'"
        result = self.handler.query(query)
        assert result['type'] is RESPONSE_TYPE.TABLE
